import ray
import logging
import pyarrow as pa
from collections import defaultdict
from ray import cloudpickle
from deltacat import logs

from ray.types import ObjectRef

from deltacat.compute.compactor import PrimaryKeyIndexVersionLocator, \
    PyArrowWriteResult
from deltacat.compute.compactor.utils import primary_key_index as pki

from typing import Any, List, Tuple

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


@ray.remote(num_cpus=1, num_returns=2)
def rewrite_index(
        object_ids: List[Any],
        s3_bucket: str,
        new_primary_key_index_version_locator: PrimaryKeyIndexVersionLocator,
        max_records_per_index_file: int) -> \
        Tuple[PyArrowWriteResult, List[ObjectRef]]:

    logger.info(f"Starting rewrite primary key index task...")
    object_refs = [cloudpickle.loads(obj_id_pkl) for obj_id_pkl in object_ids]
    logger.info(f"Getting table groups object refs...")
    table_groups_list = ray.get(object_refs)
    logger.info(f"Got {len(table_groups_list)} table groups object refs...")
    hb_index_to_tables = defaultdict(list)
    for table_groups in table_groups_list:
        for hb_index, table in enumerate(table_groups):
            if table is not None:
                hb_index_to_tables[hb_index].append(table)
    logger.info(f"Running {len(hb_index_to_tables)} rewrite index rounds...")
    pki_stats = []
    for hb_index, tables in hb_index_to_tables.items():
        table = pa.concat_tables(tables)
        hb_pki_stats = pki.write_primary_key_index_files(
            table,
            new_primary_key_index_version_locator,
            s3_bucket,
            hb_index,
            max_records_per_index_file,
        )
        pki_stats.append(hb_pki_stats)
    logger.info(f"Finished rewrite primary key index task...")
    return PyArrowWriteResult.union(pki_stats), object_refs



import logging,time
import ray
import pyarrow as pa

from collections import defaultdict

from deltacat.compute.compactor.steps.dedupe import DedupeTaskIndexWithObjectId, \
    DeltaFileLocatorToRecords
from itertools import chain, repeat

from pyarrow import compute as pc

from ray import cloudpickle

from deltacat import logs
from deltacat.storage import Delta, DeltaLocator, Partition, PartitionLocator, \
    interface as unimplemented_deltacat_storage
from deltacat.compute.compactor import MaterializeResult, PyArrowWriteResult, \
    RoundCompletionInfo
from deltacat.compute.compactor.utils import system_columns as sc
from deltacat.types.media import ContentType, DELIMITED_TEXT_CONTENT_TYPES
from typing import List, Tuple, Optional

from deltacat.utils.pyarrow import ReadKwargsProviderPyArrowSchemaOverride

from deltacat.types.tables import TABLE_CLASS_TO_SIZE_FUNC
from deltacat.utils.pyarrow import ReadKwargsProviderPyArrowCsvPureUtf8

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


@ray.remote(num_cpus=0.5)
def materialize(
        source_partition_locator: PartitionLocator,
        round_completion_info: Optional[RoundCompletionInfo],
        partition: Partition,
        mat_bucket_index: int,
        dedupe_task_idx_and_obj_id_tuples: List[DedupeTaskIndexWithObjectId],
        max_records_per_output_file: int,
        compacted_file_content_type: ContentType,
        schema: Optional[pa.Schema] = None,
        deltacat_storage=unimplemented_deltacat_storage) -> MaterializeResult:

    logger.info(f"Starting materialize task...")
    dedupe_task_idx_and_obj_ref_tuples = [
        (
            t[0],
            cloudpickle.loads(t[1])
        ) for t in dedupe_task_idx_and_obj_id_tuples
    ]
    logger.info(f"Resolved materialize task obj refs...")
    dedupe_task_indices, obj_refs = zip(
        *dedupe_task_idx_and_obj_ref_tuples
    )
    # this depends on `ray.get` result order matching input order, as per the
    # contract established in: https://github.com/ray-project/ray/pull/16763
    src_file_records_list = ray.get(list(obj_refs))
    all_src_file_records = defaultdict(list)
    for i, src_file_records in enumerate(src_file_records_list):
        dedupe_task_idx = dedupe_task_indices[i]
        for src_dfl, record_numbers in src_file_records.items():
            all_src_file_records[src_dfl].append(
                (record_numbers, repeat(dedupe_task_idx, len(record_numbers)))
            )
    manifest_cache = {}
    compacted_tables = []
    for src_dfl in sorted(all_src_file_records.keys()):
        record_numbers_dd_task_idx_tpl_list: List[Tuple[DeltaFileLocatorToRecords, repeat]] = \
            all_src_file_records[src_dfl]
        record_numbers_tpl, dedupe_task_idx_iter_tpl = zip(
            *record_numbers_dd_task_idx_tpl_list
        )
        is_src_partition_file_np = src_dfl.is_source_delta
        src_stream_position_np = src_dfl.stream_position
        src_file_idx_np = src_dfl.file_index
        src_file_partition_locator = source_partition_locator \
            if is_src_partition_file_np \
            else round_completion_info.compacted_delta_locator.partition_locator
        delta_locator = DeltaLocator.of(
            src_file_partition_locator,
            src_stream_position_np.item(),
        )
        dl_digest = delta_locator.digest()

        manifest = manifest_cache.setdefault(
            dl_digest,
            deltacat_storage.get_delta_manifest(delta_locator),
        )

        read_kwargs_provider = None
        # for delimited text output, disable type inference to prevent
        # unintentional type-casting side-effects and improve performance
        if compacted_file_content_type in DELIMITED_TEXT_CONTENT_TYPES:
            read_kwargs_provider = ReadKwargsProviderPyArrowCsvPureUtf8()
        # enforce a consistent schema if provided, when reading files into PyArrow tables
        elif schema is not None:
            read_kwargs_provider = ReadKwargsProviderPyArrowSchemaOverride(
                schema=schema)
        pa_table = deltacat_storage.download_delta_manifest_entry(
            Delta.of(delta_locator, None, None, None, manifest),
            src_file_idx_np.item(),
            file_reader_kwargs=read_kwargs_provider,
        )
        mask_pylist = list(repeat(False, len(pa_table)))
        record_numbers = chain.from_iterable(record_numbers_tpl)
        for record_number in record_numbers:
            mask_pylist[record_number] = True
        mask = pa.array(mask_pylist)
        compacted_table = pa_table.filter(mask)

        # appending, sorting, taking, and dropping has 2-3X latency of a
        # single filter on average, and thus provides better average
        # performance than repeatedly filtering the table in dedupe task index
        # order
        dedupe_task_indices = chain.from_iterable(dedupe_task_idx_iter_tpl)
        compacted_table = sc.append_dedupe_task_idx_col(
            compacted_table,
            dedupe_task_indices,
        )
        pa_sort_keys = [(sc._DEDUPE_TASK_IDX_COLUMN_NAME, "ascending")]
        compacted_table = compacted_table.take(
            pc.sort_indices(compacted_table, sort_keys=pa_sort_keys),
        )
        compacted_table = compacted_table.drop(
            [sc._DEDUPE_TASK_IDX_COLUMN_NAME]
        )
        compacted_tables.append(compacted_table)

    # TODO (pdames): save memory by writing output files eagerly whenever
    #  len(compacted_table) >= max_records_per_output_file (but don't write
    #  partial slices from the compacted_table remainder every time!)
    compacted_table = pa.concat_tables(compacted_tables)
    if compacted_file_content_type in DELIMITED_TEXT_CONTENT_TYPES:
        # convert to pandas since pyarrow doesn't support custom delimiters
        # and doesn't support utf-8 conversion of all types (e.g. Decimal128)
        # TODO (pdames): compare performance to pandas-native materialize path
        df = compacted_table.to_pandas(
            split_blocks=True,
            self_destruct=True,
        )
        del compacted_table
        compacted_table = df
    delta = deltacat_storage.stage_delta(
        compacted_table,
        partition,
        max_records_per_entry=max_records_per_output_file,
        content_type=compacted_file_content_type,
    )
    manifest = delta.manifest
    manifest_records = manifest.meta.record_count
    assert(manifest_records == len(compacted_table),
           f"Unexpected Error: Materialized delta manifest record count "
           f"({manifest_records}) does not equal compacted table record count "
           f"({len(compacted_table)})")
    materialize_result = MaterializeResult.of(
        delta,
        mat_bucket_index,
        # TODO (pdames): Generalize WriteResult to contain in-memory-table-type
        #  and in-memory-table-bytes instead of tight coupling to paBytes
        PyArrowWriteResult.of(
            len(manifest.entries),
            TABLE_CLASS_TO_SIZE_FUNC[type(compacted_table)](compacted_table),
            manifest.meta.content_length,
            len(compacted_table),
        ),
    )
    logger.info(f"Materialize result: {materialize_result}")
    logger.info(f"Finished materialize task...")
    return materialize_result



import logging
import pyarrow as pa
import ray
import time
import pyarrow.compute as pc
import numpy as np
from deltacat.compute.compactor.utils.system_columns import get_minimal_hb_schema
from deltacat.utils.pyarrow import ReadKwargsProviderPyArrowSchemaOverride
from ray import cloudpickle
from ray.types import ObjectRef

from deltacat import logs
from collections import defaultdict
from itertools import repeat
from deltacat.storage import DeltaType
from deltacat.compute.compactor import SortKey, SortOrder, \
    RoundCompletionInfo, PrimaryKeyIndexVersionLocator, DeltaFileEnvelope, \
    DeltaFileLocator, PyArrowWriteResult
from deltacat.compute.compactor.utils import system_columns as sc, \
    primary_key_index as pki

from typing import Any, Dict, List, Optional, Tuple
from ray.util.scheduling_strategies import PlacementGroupSchedulingStrategy

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


MaterializeBucketIndex = int
DeltaFileLocatorToRecords = Dict[DeltaFileLocator, np.ndarray]
DedupeTaskIndex, PickledObjectRef = int, str
DedupeTaskIndexWithObjectId = Tuple[DedupeTaskIndex, PickledObjectRef]
DedupeResult = Tuple[
    Dict[MaterializeBucketIndex, DedupeTaskIndexWithObjectId],
    List[ObjectRef[DeltaFileLocatorToRecords]],
    PyArrowWriteResult
]


def union_primary_key_indices(
        s3_bucket: str,
        round_completion_info: RoundCompletionInfo,
        hash_bucket_index: int,
        df_envelopes_list: List[List[DeltaFileEnvelope]]) -> pa.Table:

    logger.info(f"Reading dedupe input for {len(df_envelopes_list)} "
                f"delta file envelope lists...")
    # read compacted input parquet files first
    # (which implicitly have older stream positions than deltas)
    hb_tables = []
    if round_completion_info:
        tables = pki.download_hash_bucket_entries(
            s3_bucket,
            hash_bucket_index,
            round_completion_info.primary_key_index_version_locator,
            # Enforce consistent column ordering by reading from a schema, to prevent schema mismatch errors
            file_reader_kwargs_provider=ReadKwargsProviderPyArrowSchemaOverride(schema=get_minimal_hb_schema())
        )
        if tables:
            prev_compacted_delta_stream_pos = round_completion_info\
                .compacted_delta_locator \
                .stream_position
            if prev_compacted_delta_stream_pos is None:
                raise ValueError(f"Unexpected Error: No previous compacted "
                                 f"delta stream position found in round "
                                 f"completion info: {round_completion_info}")
            prior_pk_index_table = pa.concat_tables(tables)
            prior_pk_index_table = sc.append_stream_position_column(
                prior_pk_index_table,
                repeat(
                    prev_compacted_delta_stream_pos,
                    len(prior_pk_index_table),
                ),
            )
            prior_pk_index_table = sc.append_delta_type_col(
                prior_pk_index_table,
                repeat(
                    sc.delta_type_to_field(DeltaType.UPSERT),
                    len(prior_pk_index_table),
                )
            )
            prior_pk_index_table = sc.append_is_source_col(
                prior_pk_index_table,
                repeat(
                    False,
                    len(prior_pk_index_table),
                )
            )
            hb_tables.append(prior_pk_index_table)

    # sort by delta file stream position now instead of sorting every row later
    df_envelopes = [d for dfe_list in df_envelopes_list for d in dfe_list]
    df_envelopes = sorted(
        df_envelopes,
        key=lambda df: (df.stream_position, df.file_index),
        reverse=False,  # ascending
    )
    for df_envelope in df_envelopes:
        hb_tables.append(sc.project_delta_file_metadata_on_table(df_envelope))

    hb_table = pa.concat_tables(hb_tables)

    return hb_table


def drop_duplicates_by_primary_key_hash(table: pa.Table) -> pa.Table:
    value_to_last_row_idx = {}
    row_idx = 0
    pk_op_chunk_iter = zip(
        sc.pk_hash_column(table).iterchunks(),
        sc.delta_type_column(table).iterchunks(),
    )
    for (pk_chunk, op_chunk) in pk_op_chunk_iter:
        pk_op_val_iter = zip(
            pk_chunk.to_numpy(zero_copy_only=False),
            op_chunk.to_numpy(zero_copy_only=False),
        )
        for (pk_val, op_val) in pk_op_val_iter:
            # operation type is True for `UPSERT` and False for `DELETE`
            if op_val:
                # UPSERT this row
                value_to_last_row_idx[pk_val] = row_idx
            else:
                # DELETE this row
                value_to_last_row_idx.pop(pk_val, None)
            row_idx += 1
    return table.take(list(value_to_last_row_idx.values()))


def write_new_primary_key_index(
        s3_bucket: str,
        new_primary_key_index_version_locator: PrimaryKeyIndexVersionLocator,
        max_rows_per_index_file: int,
        max_rows_per_mat_file: int,
        num_materialize_buckets: int,
        dedupe_task_index: int,
        deduped_tables: List[Tuple[int, pa.Table]],
        row_counts: Dict[int, Dict[Tuple[np.bool_, np.int64, np.int32],
                                   Dict[int, int]]]) -> PyArrowWriteResult:

    logger.info(f"Writing new deduped primary key index: "
                f"{new_primary_key_index_version_locator}")
    # TODO (pdames): move to RecordCountsPendingMaterialize.finalize()?
    file_idx = 0
    prev_file_idx = 0
    dest_file_indices = defaultdict(
        lambda: defaultdict(
            lambda: defaultdict(int)
        )
    )
    dest_file_row_indices = defaultdict(
        lambda: defaultdict(
            lambda: defaultdict(int)
        )
    )
    for mat_bucket in sorted(row_counts.keys()):
        mat_bucket_row_idx = 0
        sorted_src_dfls = sorted(row_counts[mat_bucket].keys())
        for src_dfl in sorted_src_dfls:
            sorted_dd_tasks = sorted(row_counts[mat_bucket][src_dfl].keys())
            for dd_task_idx in sorted_dd_tasks:
                dest_file_row_indices[mat_bucket][src_dfl][dd_task_idx] = \
                    mat_bucket_row_idx % max_rows_per_mat_file
                file_idx = prev_file_idx + int(
                    mat_bucket_row_idx / max_rows_per_mat_file
                )
                dest_file_indices[mat_bucket][src_dfl][dd_task_idx] = file_idx
                row_count = row_counts[mat_bucket][src_dfl][dd_task_idx]
                mat_bucket_row_idx += row_count
        prev_file_idx = file_idx + 1

    pki_results = []
    src_dfl_row_counts = defaultdict(int)
    for hb_index, table in deduped_tables:
        is_source_col = sc.is_source_column_np(table)
        stream_pos_col = sc.stream_position_column_np(table)
        file_idx_col = sc.file_index_column_np(table)
        dest_file_idx_col = []
        dest_file_row_idx_col = []
        for row_idx in range(len(table)):
            src_dfl = DeltaFileLocator.of(
                is_source_col[row_idx],
                stream_pos_col[row_idx],
                file_idx_col[row_idx],
            )
            mat_bucket = delta_file_locator_to_mat_bucket_index(
                src_dfl,
                num_materialize_buckets,
            )
            dest_file_start_idx = \
                dest_file_indices[mat_bucket][src_dfl][dedupe_task_index]
            dest_file_row_idx_offset = src_dfl_row_counts[src_dfl] + \
                dest_file_row_indices[mat_bucket][src_dfl][dedupe_task_index]
            dest_file_idx_offset = int(
                dest_file_row_idx_offset / max_rows_per_mat_file
            )
            dest_file_idx = dest_file_start_idx + dest_file_idx_offset
            dest_file_idx_col.append(dest_file_idx)
            dest_file_row_idx = dest_file_row_idx_offset % max_rows_per_mat_file
            dest_file_row_idx_col.append(dest_file_row_idx)
            src_dfl_row_counts[src_dfl] += 1
        table = table.drop([
            sc._IS_SOURCE_COLUMN_NAME,
            sc._PARTITION_STREAM_POSITION_COLUMN_NAME,
            sc._ORDERED_FILE_IDX_COLUMN_NAME,
            sc._ORDERED_RECORD_IDX_COLUMN_NAME,
        ])
        table = sc.append_file_idx_column(table, dest_file_idx_col)
        table = sc.append_record_idx_col(table, dest_file_row_idx_col)

        hb_pki_result = pki.write_primary_key_index_files(
            table,
            new_primary_key_index_version_locator,
            s3_bucket,
            hb_index,
            max_rows_per_index_file,
        )
        pki_results.append(hb_pki_result)

    result = PyArrowWriteResult.union(pki_results)
    logger.info(f"Wrote new deduped primary key index: "
                f"{new_primary_key_index_version_locator}. Result: {result}")
    return result


def delta_file_locator_to_mat_bucket_index(
        df_locator: DeltaFileLocator,
        materialize_bucket_count: int) -> int:
    digest = df_locator.digest()
    return int.from_bytes(digest, "big") % materialize_bucket_count


@ray.remote(num_cpus=0.2)
class RecordCountsPendingMaterialize:
    def __init__(self, expected_result_count: int):
        # materialize_bucket -> src_file_id
        self.record_counts = defaultdict(
            # delta_file_locator -> dedupe task index
            lambda: defaultdict(
                # dedupe task index -> row count
                lambda: defaultdict(int)
            )
        )
        self.expected_result_count = expected_result_count
        self.actual_result_count = 0

    def add_record_counts(
            self,
            result_idx: int,
            record_counts:
            Dict[int, Dict[Tuple[np.bool_, np.int64, np.int32], int]]) -> None:
        for mat_bucket, df_locator_rows in record_counts.items():
            for df_locator, rows in df_locator_rows.items():
                self.record_counts[mat_bucket][df_locator][result_idx] += rows
        self.actual_result_count += 1

    def get_record_counts(self) -> \
            Dict[int, Dict[Tuple[np.bool_, np.int64, np.int32],
                           Dict[int, int]]]:
        return self.record_counts

    def get_expected_result_count(self) -> int:
        return self.expected_result_count

    def get_actual_result_count(self) -> int:
        return self.actual_result_count

    def is_finalized(self) -> bool:
        return self.actual_result_count == self.expected_result_count


@ray.remote(num_cpus=0.5,num_returns=3)
def dedupe(
        compaction_artifact_s3_bucket: str,
        round_completion_info: Optional[RoundCompletionInfo],
        new_primary_key_index_version_locator: PrimaryKeyIndexVersionLocator,
        object_ids: List[Any],
        sort_keys: List[SortKey],
        max_records_per_index_file: int,
        max_records_per_materialized_file: int,
        num_materialize_buckets: int,
        dedupe_task_index: int,
        delete_old_primary_key_index: bool,
        record_counts_pending_materialize: RecordCountsPendingMaterialize) -> DedupeResult:

    logger.info(f"Starting dedupe task...")
    # TODO (pdames): mitigate risk of running out of memory here in cases of
    #  severe skew of primary key updates in deltas
    src_file_records_obj_refs = [
        cloudpickle.loads(obj_id_pkl) for obj_id_pkl in object_ids]
    logger.info(f"Getting delta file envelope groups object refs...")
    delta_file_envelope_groups_list = ray.get(src_file_records_obj_refs)
    hb_index_to_delta_file_envelopes_list = defaultdict(list)
    for delta_file_envelope_groups in delta_file_envelope_groups_list:
        for hb_idx, dfes in enumerate(delta_file_envelope_groups):
            if dfes is not None:
                hb_index_to_delta_file_envelopes_list[hb_idx].append(dfes)
    src_file_id_to_row_indices = defaultdict(list)
    deduped_tables = []
    logger.info(f"Running {len(hb_index_to_delta_file_envelopes_list)} "
                f"dedupe rounds...")
    for hb_idx, dfe_list in hb_index_to_delta_file_envelopes_list.items():
        table = union_primary_key_indices(
            compaction_artifact_s3_bucket,
            round_completion_info,
            hb_idx,
            dfe_list,
        )
        logger.info(f"Dedupe round input record count: {len(table)}")

        # sort by sort keys
        if len(sort_keys):
            # TODO (pdames): convert to O(N) dedupe w/ sort keys
            sort_keys.extend([
                SortKey.of(
                    sc._PARTITION_STREAM_POSITION_COLUMN_NAME,
                    SortOrder.ASCENDING
                ),
                SortKey.of(
                    sc._ORDERED_FILE_IDX_COLUMN_NAME,
                    SortOrder.ASCENDING
                ),
            ])
            table = table.take(pc.sort_indices(table, sort_keys=sort_keys))

        # drop duplicates by primary key hash column
        table = drop_duplicates_by_primary_key_hash(table)
        table = table.drop([sc._DELTA_TYPE_COLUMN_NAME])
        logger.info(f"Dedupe round output record count: {len(table)}")

        deduped_tables.append((hb_idx, table))

        stream_position_col = sc.stream_position_column_np(table)
        file_idx_col = sc.file_index_column_np(table)
        row_idx_col = sc.record_index_column_np(table)
        is_source_col = sc.is_source_column_np(table)
        for row_idx in range(len(table)):
            src_dfl = DeltaFileLocator.of(
                is_source_col[row_idx],
                stream_position_col[row_idx],
                file_idx_col[row_idx],
            )
            # TODO(pdames): merge contiguous record number ranges
            src_file_id_to_row_indices[src_dfl].append(row_idx_col[row_idx])

    logger.info(f"Finished all dedupe rounds...")
    mat_bucket_to_src_file_record_count = defaultdict(dict)
    mat_bucket_to_src_file_records: Dict[MaterializeBucketIndex, DeltaFileLocatorToRecords] = defaultdict(dict)
    for src_dfl, src_row_indices in src_file_id_to_row_indices.items():
        mat_bucket = delta_file_locator_to_mat_bucket_index(
            src_dfl,
            num_materialize_buckets,
        )
        mat_bucket_to_src_file_records[mat_bucket][src_dfl] = np.array(
            src_row_indices,
        )
        mat_bucket_to_src_file_record_count[mat_bucket][src_dfl] = \
            len(src_row_indices)

    mat_bucket_to_dd_idx_obj_id: Dict[MaterializeBucketIndex, DedupeTaskIndexWithObjectId] = {}
    src_file_records_obj_refs: List[ObjectRef[DeltaFileLocatorToRecords]] = []
    for mat_bucket, src_file_records in mat_bucket_to_src_file_records.items():
        object_ref = ray.put(src_file_records)
        src_file_records_obj_refs.append(object_ref)
        mat_bucket_to_dd_idx_obj_id[mat_bucket] = (
            dedupe_task_index,
            cloudpickle.dumps(object_ref),
        )
    logger.info(f"Count of materialize buckets with object refs: "
                f"{len(mat_bucket_to_dd_idx_obj_id)}")

    record_counts_pending_materialize.add_record_counts.remote(
        dedupe_task_index,
        mat_bucket_to_src_file_record_count,
    )

    # wait for all dedupe tasks to reach this point before continuing
    logger.info(
        f"Waiting for all dedupe tasks to finish writing record counts...")
    finalized = False
    while not finalized:
        finalized = ray.get(
            record_counts_pending_materialize.is_finalized.remote()
        )
        time.sleep(0.25)
    record_counts = ray.get(
        record_counts_pending_materialize.get_record_counts.remote()
    )

    write_pki_result: PyArrowWriteResult = write_new_primary_key_index(
        compaction_artifact_s3_bucket,
        new_primary_key_index_version_locator,
        max_records_per_index_file,
        max_records_per_materialized_file,
        num_materialize_buckets,
        dedupe_task_index,
        deduped_tables,
        record_counts,
    )

    if delete_old_primary_key_index:
        pki.delete_primary_key_index_version(
            compaction_artifact_s3_bucket,
            round_completion_info.primary_key_index_version_locator,
        )
    logger.info(f"Finished dedupe task...")
    return mat_bucket_to_dd_idx_obj_id, \
        src_file_records_obj_refs, \
        write_pki_result



from deltacat.aws.redshift.model.manifest import Manifest, ManifestAuthor, \
    ManifestEntry, ManifestMeta, ManifestEntryList

__all__ = [
    "Manifest",
    "ManifestAuthor",
    "ManifestEntry",
    "ManifestMeta",
    "ManifestEntryList",
]



import os
import subprocess
import ray
import errno
import logging

from deltacat import logs
from tenacity import retry, stop_after_attempt
from typing import Any
from deltacat.compute.metastats.utils.constants import WORKER_NODE_OBJECT_STORE_MEMORY_RESERVE_RATIO, MAX_WORKER_MULTIPLIER
from tenacity import Retrying, stop_after_attempt, wait_fixed, RetryError
from subprocess import run, PIPE

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))

RAY_DOWN_DEFAULT_RETRY_ATTEMPTS = 3


def run_cmd_exit_code(cmd: str) -> int:
    logger.info(f"running command {cmd}")
    exit_code = int(os.system(cmd))
    logger.info(f"Got {exit_code} when running {cmd}")


def run_cmd_with_retry(cmd: str) -> None:
    retrying = Retrying(
        wait=wait_fixed(2),
        stop=stop_after_attempt(RAY_DOWN_DEFAULT_RETRY_ATTEMPTS)
    )
    try:
        retrying(
           run_cmd_exit_code(cmd)
        )
    except RetryError:
        logger.info(f"{cmd} failed after {RAY_DOWN_DEFAULT_RETRY_ATTEMPTS} retries.")


def run_cmd(cmd: str) -> None:
    result = run(cmd, shell=True, capture_output=True)
    exit_code = int(result.returncode)
    assert exit_code == 0, f"`{cmd}` failed. Exit code: {exit_code} " \
                           f"Error Trace: {result.stderr}"


def ray_up(cluster_cfg: str) -> None:
    logger.info(f"Starting Ray cluster '{cluster_cfg}'")
    run_cmd(f"ray up {cluster_cfg} -y --no-config-cache --no-restart")
    logger.info(f"Started Ray cluster '{cluster_cfg}'")


def ray_down(cluster_cfg: str) -> None:
    logger.info(f"Destroying Ray cluster '{cluster_cfg}'")
    run_cmd_with_retry(f"ray down {cluster_cfg} -y")
    logger.info(f"Destroyed Ray cluster '{cluster_cfg}'")


def clean_up_cluster_cfg_file(cluster_cfg) -> None:
    logger.info(f"Removing stats cluster config at: '{cluster_cfg}'")
    run_cmd(f"rm -f {cluster_cfg}")
    logger.info(f"Removed stats cluster config at: '{cluster_cfg}'")


def get_head_node_ip(cluster_cfg: str) -> str:
    logger.info(f"Getting Ray cluster head node IP for '{cluster_cfg}'")
    proc = subprocess.run(
        f"ray get-head-ip {cluster_cfg}",
        shell=True,
        capture_output=True,
        text=True,
        check=True)
    # the head node IP should be the last line printed to stdout
    head_node_ip = proc.stdout.splitlines()[-1]
    logger.info(f"Ray cluster head node IP for '{cluster_cfg}': {head_node_ip}")
    return head_node_ip


def ray_init(host, port) -> Any:
    ray_init_uri = f"ray://{host}:{port}"
    logger.info(f"Connecting Ray Client to '{ray_init_uri}'")
    client = ray.init(ray_init_uri, allow_multiple=True)
    logger.info(f"Connected Ray Client to '{ray_init_uri}'")
    return client


def replace_cluster_cfg_vars(
        partition_canonical_string: str,
        trace_id: str,
        file_path: str,
        min_workers: int,
        head_type: str,
        worker_type: str,
        head_object_store_memory_pct: int,
        worker_object_store_memory_pct: int) -> str:

    head_object_store_memory_pct = head_object_store_memory_pct if not None else 30
    worker_object_store_memory_pct = WORKER_NODE_OBJECT_STORE_MEMORY_RESERVE_RATIO * 100

    max_workers = int(min_workers * MAX_WORKER_MULTIPLIER)
    with open(file_path, "r+") as file:
        contents = file.read().replace("{{use-internal-ips}}", "True")
        contents = contents.replace("{{partition_canonical_string}}", partition_canonical_string)
        contents = contents.replace("{{trace_id}}", trace_id)
        contents = contents.replace("{{min-workers}}", str(min_workers))
        contents = contents.replace("{{max-workers}}", str(max_workers))
        contents = contents.replace("{{head-instance-type}}", head_type)
        contents = contents.replace("{{worker-instance-type}}", worker_type)
        contents = contents.replace(
            "{{head-object-store-memory-pct}}",
            str(head_object_store_memory_pct))
        contents = contents.replace(
            "{{worker-object-store-memory-pct}}",
            str(worker_object_store_memory_pct))
    partition_id = partition_canonical_string.split("|")[-1]
    out_file_name = f"{trace_id}-{partition_id}.{os.path.basename(file_path)}"
    out_file_dir = os.path.join(os.path.dirname(file_path), "tmp")
    out_file_path = os.path.join(out_file_dir, out_file_name)
    try:
        os.makedirs(os.path.dirname(out_file_path))
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise
    with open(out_file_path, "w") as output:
        output.write(contents)
    return out_file_path



import time
from collections import Counter
from typing import Any, Callable, Tuple


def invoke_with_perf_counter(
        counter: Counter,
        counter_key: Any,
        func: Callable,
        *args,
        **kwargs) -> Tuple[Any, float]:

    start = time.perf_counter()
    result = func(*args, **kwargs)
    stop = time.perf_counter()
    latency = stop - start
    counter[counter_key] += latency
    return result, latency


def timed_invocation(
        func: Callable,
        *args,
        **kwargs) -> Tuple[Any, float]:

    start = time.perf_counter()
    result = func(*args, **kwargs)
    stop = time.perf_counter()
    return result, stop - start



import numpy as np
import pyarrow as pa

from fsspec import AbstractFileSystem

from ray.data.datasource import BlockWritePathProvider

from deltacat.types.media import ContentType
from deltacat.utils import pyarrow as pa_utils
from deltacat.utils import pandas as pd_utils
from deltacat.utils.common import ReadKwargsProvider

from typing import List, Optional


def slice_ndarray(
        np_array: np.ndarray,
        max_len: Optional[int]) -> List[np.ndarray]:
    """
    Iteratively creates max_len slices from the first dimension of an ndarray.
    """
    if max_len is None:
        return [np_array]

    # Slice along the first dimension of the ndarray.
    return [np_array[i:i + max_len] for i in range(0, len(np_array), max_len)]


def s3_file_to_ndarray(
        s3_url: str,
        content_type: str,
        content_encoding: str,
        column_names: Optional[List[str]] = None,
        include_columns: Optional[List[str]] = None,
        pd_read_func_kwargs_provider: Optional[ReadKwargsProvider] = None,
        **s3_client_kwargs) -> np.ndarray:
    # TODO: Compare perf to s3 -> pyarrow -> pandas [Series/DataFrame] -> numpy
    dataframe = pd_utils.s3_file_to_dataframe(
        s3_url,
        content_type,
        content_encoding,
        column_names,
        include_columns,
        pd_read_func_kwargs_provider,
        **s3_client_kwargs
    )
    return dataframe.to_numpy()


def ndarray_size(np_array: np.ndarray) -> int:
    return np_array.nbytes


def ndarray_to_file(
        np_array: np.ndarray,
        path: str,
        file_system: AbstractFileSystem,
        block_path_provider: BlockWritePathProvider,
        content_type: str = ContentType.PARQUET.value,
        **kwargs) -> None:
    """
    Writes the given Numpy ndarray to a file.
    """

    # PyArrow only supports 1D ndarrays, so convert to list of 1D arrays
    np_arrays = [array for array in np_array]
    pa_utils.table_to_file(
        pa.table({"data": np_arrays}),
        path,
        file_system,
        block_path_provider,
        content_type,
        **kwargs
    )



import logging
import json
import pyarrow
import ray

from deltacat import LocalTable, TableType
from deltacat.storage import Delta
from deltacat.compute.compactor import DeltaAnnotated
from deltacat.aws import s3u as s3_utils
from deltacat.utils.common import sha1_hexdigest
from deltacat.storage import interface as unimplemented_deltacat_storage
from deltacat.compute.metastats.model.partition_stats_dict import PartitionStats
from deltacat.compute.stats.models.delta_stats_cache_result import DeltaStatsCacheResult
from deltacat.compute.stats.models.delta_column_stats import DeltaColumnStats
from deltacat.compute.stats.models.delta_stats import DeltaStats, DeltaStatsCacheMiss
from deltacat.compute.stats.models.stats_result import StatsResult

from typing import Dict, List, Optional, Any
from collections import defaultdict
from deltacat import logs
logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def cache_inflation_rate_data_for_delta_stats_ready(delta_stats_processed_list, inflation_rate_stats_s3_url,
                                                     deltacat_storage):
    meta_stats_processed_list: Dict[int, int] = {}

    for key, value in delta_stats_processed_list.items():
        delta_locator = value.column_stats[0].manifest_stats.delta_locator
        delta_meta_count = 0
        manifest = deltacat_storage.get_delta_manifest(delta_locator)
        delta = Delta.of(delta_locator, None, None, None, manifest)
        for entry in delta.manifest.entries:
            delta_meta_count += entry.meta.content_length
        meta_stats_processed_list[delta.stream_position] = delta_meta_count

    cache_inflation_rate_res = dict()

    for key, value in delta_stats_processed_list.items():
        delta_stats_pyarrow_bytes_sum = 0
        delta_stats_row_count = 0
        for column_stats in delta_stats_processed_list[key].column_stats[0].manifest_stats.stats:
            delta_stats_row_count += column_stats.get("rowCount")
        for stats in delta_stats_processed_list[key].get("column_stats"):

            delta_stats_pyarrow_bytes_sum += stats.get("stats").get("pyarrowTableBytes")
        cache_inflation_rate_res[key] = [meta_stats_processed_list[key], delta_stats_row_count,
                                         delta_stats_pyarrow_bytes_sum]

    if inflation_rate_stats_s3_url:
        logger.warning(
            f"reading previous inflation rate stats from: {inflation_rate_stats_s3_url}")

        result = s3_utils.download(inflation_rate_stats_s3_url, fail_if_not_found=False)

        prev_inflation_rate_stats = dict()
        if result:
            json_str = result["Body"].read().decode("utf-8")
            prev_inflation_rate_stats_read = json.loads(json_str)
            prev_inflation_rate_stats = prev_inflation_rate_stats_read if prev_inflation_rate_stats_read else dict()
            logger.debug(f"read stats completion info: {prev_inflation_rate_stats_read}")
        logger.debug(
            f"writing inflation rate info to S3: {inflation_rate_stats_s3_url}")
        prev_inflation_rate_stats.update(cache_inflation_rate_res)
        logger.debug(f"writing current inflation rate info to S3: {prev_inflation_rate_stats}")
        s3_utils.upload(
            inflation_rate_stats_s3_url,
            json.dumps(prev_inflation_rate_stats)
        )
    else:
        logger.warning(f"No valid s3 url received to cache inflation rate stats, got {inflation_rate_stats_s3_url}")


def read_cached_partition_stats(partition_canonical_string: str, stat_results_s3_bucket: str):
    partition_stats_url = get_partition_stats_s3_url(partition_canonical_string, stat_results_s3_bucket)
    logger.info(
        f"reading partition stats completion file from: {partition_stats_url}")

    result = s3_utils.download(partition_stats_url, fail_if_not_found=False)
    delta_stats_cache_res_map: Dict[int, List[DeltaStatsCacheResult]] = {}
    if result:
        json_str = result["Body"].read().decode("utf-8")
        partition_stats_str = json.loads(json_str)
        delta_stats_cache_res_map = get_delta_stats_from_partition_stats(partition_stats_str)

    return delta_stats_cache_res_map


def get_partition_stats_s3_url(partition_canonical_string: str, stat_results_s3_bucket: str):
    stats_partition_canonical_string = f"{partition_canonical_string}"
    stats_partition_hexdigest = sha1_hexdigest(stats_partition_canonical_string.encode("utf-8"))
    base_path = s3_utils.parse_s3_url(stat_results_s3_bucket).url

    return f"{base_path}/{stats_partition_hexdigest}.json"


def get_delta_stats_from_partition_stats(partition_stats_str: str):

    partition_stats = PartitionStats.build_from_dict(partition_stats_str)

    found_columns_stats_map: Dict[int, List[DeltaStatsCacheResult]] = {}
    for stream_position, delta_stats in partition_stats.delta_stats.items():
        found_columns_stats: List[DeltaColumnStats] = []
        missed_columns: List[str] = []
        for cs in delta_stats.column_stats:
            if cs.manifest_stats:
                found_columns_stats.append(cs)
            else:
                missed_columns.append(cs.column)

        delta_locator = delta_stats.column_stats[0].manifest_stats.delta_locator
        found_stats: Optional[DeltaStats] = DeltaStats.of(found_columns_stats) if found_columns_stats else None
        missed_stats: Optional[DeltaStatsCacheMiss] = DeltaStatsCacheMiss(missed_columns, delta_locator) \
            if missed_columns else None
        delta_stats_cache_res = DeltaStatsCacheResult.of(found_stats, missed_stats)
        found_columns_stats_map[int(stream_position)] = delta_stats_cache_res
    return found_columns_stats_map


def cache_partition_stats_to_s3(stat_results_s3_bucket, delta_stream_range_stats, partition_canonical_string):
    partition_stats = PartitionStats.of(delta_stream_range_stats, partition_canonical_string)
    logger.info(
        f"writing partition stats completion for {partition_canonical_string}")
    partition_stats_completion_file_s3_url = get_partition_stats_s3_url(
        partition_canonical_string,
        stat_results_s3_bucket
    )
    s3_utils.upload(
        partition_stats_completion_file_s3_url,
        str(json.dumps(partition_stats))
    )
    logger.debug(
        f"stats completion file written to: {partition_stats_completion_file_s3_url}")


@ray.remote
def collect_stats_by_columns(delta_annotated: DeltaAnnotated,
                              columns_to_compute: Optional[List[str]] = None,
                              deltacat_storage=unimplemented_deltacat_storage) -> Dict[str, Any]:
    """Materializes one manifest entry at a time to save memory usage and calculate stats from each of its columns.

    Args:
        delta: A delta object to calculate stats for
        columns_to_compute: Columns to calculate stats for. If not provided, all columns are considered.
        deltacat_storage: Client implementation of the DeltaCAT storage interface

    Returns:
        A delta wide stats container
    """
    total_tables_size = 0

    # Mapping of column_name -> [stats_file_idx_1, stats_file_idx_2, ... stats_file_idx_n]
    column_stats_map = defaultdict(lambda: [[None, None]] * len(delta_annotated["manifest"].get("entries")))
    src_da_entries = delta_annotated["manifest"].get("entries")
    manifest_annotations = delta_annotated["annotations"]
    for file_idx, manifest in enumerate(src_da_entries):
        entry_pyarrow_table: LocalTable = \
            deltacat_storage.download_delta_manifest_entry(delta_annotated, file_idx, TableType.PYARROW, columns_to_compute, equivalent_table_types="uncompacted")
        assert isinstance(entry_pyarrow_table, pyarrow.Table), \
            f"Stats collection is only supported for PyArrow tables, but received a table of " \
            f"type '{type(entry_pyarrow_table)}' for manifest entry {file_idx} of delta: {delta_annotated.locator}."
        total_tables_size += entry_pyarrow_table.nbytes
        if not columns_to_compute:
            columns_to_compute = entry_pyarrow_table.column_names

        for column_idx, pyarrow_column in enumerate(entry_pyarrow_table.columns):
            column_name = columns_to_compute[column_idx]
            origin_delta_stream_position = manifest_annotations[file_idx][-1]
            column_stats_map[column_name][file_idx] = [StatsResult.of(len(pyarrow_column), pyarrow_column.nbytes),
                                                       origin_delta_stream_position]

    return column_stats_map


import ray
from typing import Dict, Set, Tuple, List, Optional

from deltacat.compute.stats.models.delta_stats import DeltaStats
from ray.types import ObjectRef

from deltacat.compute.stats.models.delta_stats_cache_result import DeltaStatsCacheResult
from deltacat.compute.stats.models.stats_result import StatsResult
from deltacat.compute.stats.types import StatsType
from deltacat.compute.stats.utils.io import read_cached_delta_stats, cache_delta_column_stats, get_delta_stats, \
    get_deltas_from_range
from deltacat.compute.stats.utils.intervals import merge_intervals, DeltaRange

from deltacat.storage import PartitionLocator, DeltaLocator, Delta
from deltacat.storage import interface as unimplemented_deltacat_storage


# TODO (ricmiyam): Decouple DeltaCAT from S3-based paths
# TODO (ricmiyam): Determine cache eviction policy


def collect(
        source_partition_locator: PartitionLocator,
        delta_stream_position_range_set: Optional[Set[DeltaRange]] = None,
        columns: Optional[List[str]] = None,
        stat_results_s3_bucket: Optional[str] = None,
        deltacat_storage=unimplemented_deltacat_storage) -> Dict[int, DeltaStats]:
    """Collects statistics on deltas, given a set of delta stream position ranges.

    Example:
        >>> collect(locator, set((1, 5), (4, 8), (13, 16)))
        {
            1: DeltaStats(),  # DeltaStats for stream positions 1 - 8
            13: DeltaStats()  # DeltaStats for stream positions 13 - 16
        }

    Args:
        source_partition_locator: Reference to the partition locator tied to the given delta stream positions
        delta_stream_position_range_set: A set of intervals with an int type representing finite,
            closed bounded values, and a None type representing unbounded infinity.
        columns: Columns can be optionally included to collect stats on specific columns.
            By default, all columns will be calculated.
        stat_results_s3_bucket: Used as a cache file storage for computed delta stats
        deltacat_storage: Client implementation of the DeltaCAT storage interface

    Returns:
        A mapping of stream positions to their corresponding delta stats.
    """
    if delta_stream_position_range_set is None:
        delta_stream_position_range_set = {(None, None)}
    delta_stream_range_stats: Dict[int, DeltaStats] = {}
    delta_range_lookup_pending: List[ObjectRef[List[Delta]]] = []

    if not columns:
        columns = deltacat_storage.get_table_version_column_names(source_partition_locator.namespace,
                                                                  source_partition_locator.table_name,
                                                                  source_partition_locator.table_version)
    for range_pair in merge_intervals(delta_stream_position_range_set):
        begin, end = range_pair
        promise: ObjectRef[List[Delta]] = get_deltas_from_range.remote(source_partition_locator, begin, end,
                                                                       deltacat_storage)
        delta_range_lookup_pending.append(promise)

    delta_list_by_ranges: List[List[Delta]] = ray.get(delta_range_lookup_pending)
    deltas = [delta for delta_list in delta_list_by_ranges for delta in delta_list]

    delta_stats_processed_list: List[DeltaStats] = _collect_stats_from_deltas(deltas,
                                                                              columns,
                                                                              stat_results_s3_bucket,
                                                                              deltacat_storage)

    for delta_column_stats in delta_stats_processed_list:
        assert len(delta_column_stats.column_stats) > 0, \
            f"Expected columns of `{delta_column_stats}` to be non-empty"
        stream_position = delta_column_stats.column_stats[0].manifest_stats.delta_locator.stream_position
        delta_stream_range_stats[stream_position] = delta_column_stats

    return delta_stream_range_stats


def collect_from_deltas(
        deltas: List[Delta],
        stat_types: Set[StatsType],
        columns: Optional[List[str]] = None,
        stat_results_s3_bucket: Optional[str] = None,
        deltacat_storage=unimplemented_deltacat_storage) -> StatsResult:
    """
    Variant of the `collect` function that takes a list of deltas and computes
    the aggregate of all the delta stats.
    """
    if columns is None and deltas:
        delta_locator: DeltaLocator = deltas[0].locator
        columns = deltacat_storage.get_table_version_column_names(delta_locator.namespace,
                                                                  delta_locator.table_name,
                                                                  delta_locator.table_version)

    delta_stats_processed_list: List[DeltaStats] = \
        _collect_stats_from_deltas(deltas, columns, stat_results_s3_bucket, deltacat_storage)

    return StatsResult.merge([delta_ds.stats for delta_ds in delta_stats_processed_list], stat_types)


def _collect_stats_from_deltas(
        deltas: List[Delta],
        columns: Optional[List[str]] = None,
        stat_results_s3_bucket: Optional[str] = None,
        deltacat_storage=unimplemented_deltacat_storage) -> List[DeltaStats]:
    delta_cache_lookup_pending: List[ObjectRef[DeltaStatsCacheResult]] = []
    delta_stats_compute_pending: List[ObjectRef[DeltaStats]] = []

    for delta in deltas:
        if stat_results_s3_bucket:
            promise: ObjectRef[DeltaStatsCacheResult] = \
                read_cached_delta_stats.remote(delta, columns, stat_results_s3_bucket)
            delta_cache_lookup_pending.append(promise)
            continue

        delta_stats_compute_pending.append(get_delta_stats.remote(delta.locator, columns, deltacat_storage))

    return _process_stats(delta_cache_lookup_pending, delta_stats_compute_pending,
                          stat_results_s3_bucket, deltacat_storage)


def _process_stats(
        delta_cache_lookup_pending: List[ObjectRef[DeltaStatsCacheResult]],
        delta_stats_compute_pending: List[ObjectRef[DeltaStats]],
        stat_results_s3_bucket: Optional[str] = None,
        deltacat_storage=unimplemented_deltacat_storage) -> List[DeltaStats]:
    if stat_results_s3_bucket:
        delta_stats_processed_list: List[DeltaStats] = _resolve_pending_stats_and_cache(delta_cache_lookup_pending,
                                                                                        stat_results_s3_bucket,
                                                                                        deltacat_storage)
    else:
        delta_stats_processed_list: List[DeltaStats] = _resolve_pending_stats(delta_stats_compute_pending)

    return delta_stats_processed_list


def _resolve_pending_stats_and_cache(delta_cache_lookup_pending: List[ObjectRef[DeltaStatsCacheResult]],
                                     stat_results_s3_bucket: str,
                                     deltacat_storage) -> List[DeltaStats]:
    delta_stats_cached_list, delta_stats_pending_list = \
        _get_cached_and_pending_stats(delta_cache_lookup_pending, deltacat_storage)
    delta_stats_resolved_list: List[DeltaStats] = _resolve_pending_stats(delta_stats_pending_list)

    # Cache the stats into the file store
    delta_stats_to_cache: List[ObjectRef] = [cache_delta_column_stats.remote(stat_results_s3_bucket, dcs)
                                             for dataset_stats in delta_stats_resolved_list
                                             for dcs in dataset_stats.column_stats]
    ray.get(delta_stats_to_cache)

    return [*delta_stats_cached_list, *delta_stats_resolved_list]


def _get_cached_and_pending_stats(discover_deltas_pending: List[ObjectRef[DeltaStatsCacheResult]],
                                  deltacat_storage=unimplemented_deltacat_storage) \
        -> Tuple[List[DeltaStats], List[ObjectRef[DeltaStats]]]:
    """
    Returns a tuple of a list of delta stats fetched from the cache, and a list of Ray tasks which will
    calculate the stats for deltas on cache miss.
    """
    delta_stats_processed: List[DeltaStats] = []
    delta_stats_pending: List[ObjectRef[DeltaStats]] = []
    while discover_deltas_pending:
        ready, discover_deltas_pending = ray.wait(discover_deltas_pending)

        cached_results: List[DeltaStatsCacheResult] = ray.get(ready)
        for cached_result in cached_results:
            if cached_result.hits:
                delta_stats_processed.append(cached_result.hits)

            if cached_result.misses:
                missed_column_names: List[str] = cached_result.misses.column_names
                delta_locator: DeltaLocator = cached_result.misses.delta_locator
                delta_stats_pending.append(get_delta_stats.remote(delta_locator, missed_column_names, deltacat_storage))

    return delta_stats_processed, delta_stats_pending


def _resolve_pending_stats(delta_stats_pending_list: List[ObjectRef[DeltaStats]]) \
        -> List[DeltaStats]:
    delta_stats_processed_list: List[DeltaStats] = []
    while delta_stats_pending_list:
        ready, delta_stats_pending_list = ray.wait(delta_stats_pending_list)
        processed_stats_batch: List[DeltaStats] = ray.get(ready)
        delta_stats_processed_list.extend(processed_stats_batch)

    return delta_stats_processed_list



from deltacat.utils.common import env_integer
from typing import List


BOTO_MAX_RETRIES = env_integer("BOTO_MAX_RETRIES", 15)
TIMEOUT_ERROR_CODES: List[str] = ["ReadTimeoutError", "ConnectTimeoutError"]



import pyarrow as pa
import gzip
import bz2
import io
import logging

from functools import partial
from fsspec import AbstractFileSystem

from pyarrow import feather as paf, parquet as papq, csv as pacsv, \
    json as pajson

from ray.data.datasource import BlockWritePathProvider

from deltacat import logs
from deltacat.types.media import ContentType, ContentEncoding, \
    DELIMITED_TEXT_CONTENT_TYPES, TABULAR_CONTENT_TYPES
from deltacat.utils.common import ReadKwargsProvider, ContentTypeKwargsProvider
from deltacat.utils.performance import timed_invocation

from typing import Any, Callable, Dict, List, Optional, Iterable, Union

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


CONTENT_TYPE_TO_PA_READ_FUNC: Dict[str, Callable] = {
    ContentType.UNESCAPED_TSV.value: pacsv.read_csv,
    ContentType.TSV.value: pacsv.read_csv,
    ContentType.CSV.value: pacsv.read_csv,
    ContentType.PSV.value: pacsv.read_csv,
    ContentType.PARQUET.value: papq.read_table,
    ContentType.FEATHER.value: paf.read_table,
    # Pyarrow.orc is disabled in Pyarrow 0.15, 0.16:
    # https://issues.apache.org/jira/browse/ARROW-7811
    # ContentType.ORC.value: paorc.ContentType.ORCFile,
    ContentType.JSON.value: pajson.read_json
}


def write_feather(
        table: pa.Table,
        path: str,
        *,
        filesystem: AbstractFileSystem,
        **kwargs) -> None:

    with filesystem.open(path, "wb") as f:
        paf.write_feather(table, f, **kwargs)


def write_csv(
        table: pa.Table,
        path: str,
        *,
        filesystem: AbstractFileSystem,
        **kwargs) -> None:

    with filesystem.open(path, "wb") as f:
        # TODO (pdames): Add support for client-specified compression types.
        with pa.CompressedOutputStream(f, ContentEncoding.GZIP.value) as out:
            if kwargs.get("write_options") is None:
                # column names are kept in table metadata, so omit header
                kwargs["write_options"] = pacsv.WriteOptions(
                    include_header=False)
            pacsv.write_csv(table, out, **kwargs)


CONTENT_TYPE_TO_PA_WRITE_FUNC: Dict[str, Callable] = {
    # TODO (pdames): add support for other delimited text content types as
    #  pyarrow adds support for custom delimiters, escaping, and None value
    #  representations to pyarrow.csv.WriteOptions.
    ContentType.CSV.value: write_csv,
    ContentType.PARQUET.value: papq.write_table,
    ContentType.FEATHER.value: write_feather,
}


def content_type_to_reader_kwargs(content_type: str) -> Dict[str, Any]:
    if content_type == ContentType.UNESCAPED_TSV.value:
        return {
            "parse_options": pacsv.ParseOptions(
                delimiter="\t",
                quote_char=False),
            "convert_options": pacsv.ConvertOptions(
                null_values=[""],  # pyarrow defaults are ["", "NULL", "null"]
                strings_can_be_null=True,
            )
        }
    if content_type == ContentType.TSV.value:
        return {"parse_options": pacsv.ParseOptions(delimiter="\t")}
    if content_type == ContentType.CSV.value:
        return {"parse_options": pacsv.ParseOptions(delimiter=",")}
    if content_type == ContentType.PSV.value:
        return {"parse_options": pacsv.ParseOptions(delimiter="|")}
    if content_type in {ContentType.PARQUET.value,
                        ContentType.FEATHER.value,
                        ContentType.JSON.value}:
        return {}
    # Pyarrow.orc is disabled in Pyarrow 0.15, 0.16:
    # https://issues.apache.org/jira/browse/ARROW-7811
    # if DataTypes.ContentType.ORC:
    #   return {},
    raise ValueError(f"Unsupported content type: {content_type}")


# TODO (pdames): add deflate and snappy
ENCODING_TO_FILE_INIT: Dict[str, Callable] = {
    ContentEncoding.GZIP.value: partial(gzip.GzipFile, mode='rb'),
    ContentEncoding.BZIP2.value: partial(bz2.BZ2File, mode='rb'),
    ContentEncoding.IDENTITY.value: lambda fileobj: fileobj,
}


def slice_table(
        table: pa.Table,
        max_len: Optional[int]) -> List[pa.Table]:
    """
    Iteratively create 0-copy table slices.
    """
    if max_len is None:
        return [table]
    tables = []
    offset = 0
    records_remaining = len(table)
    while records_remaining > 0:
        records_this_entry = min(
            max_len,
            records_remaining
        )
        tables.append(table.slice(offset, records_this_entry))
        records_remaining -= records_this_entry
        offset += records_this_entry
    return tables


class ReadKwargsProviderPyArrowCsvPureUtf8(ContentTypeKwargsProvider):
    """ReadKwargsProvider impl that reads columns of delimited text files
    as UTF-8 strings (i.e. disables type inference). Useful for ensuring
    lossless reads of UTF-8 delimited text datasets and improving read
    performance in cases where type casting is not required."""
    def __init__(self, include_columns: Optional[Iterable[str]] = None):
        self.include_columns = include_columns

    def _get_kwargs(
            self,
            content_type: str,
            kwargs: Dict[str, Any]) -> Dict[str, Any]:
        if content_type in DELIMITED_TEXT_CONTENT_TYPES:
            convert_options: pacsv.ConvertOptions = \
                kwargs.get("convert_options")
            if convert_options is None:
                convert_options = pacsv.ConvertOptions()
            # read only the included columns as strings?
            column_names = self.include_columns \
                if self.include_columns else convert_options.include_columns
            if not column_names:
                # read all columns as strings?
                read_options: pacsv.ReadOptions = kwargs.get("read_options")
                if read_options and read_options.column_names:
                    column_names = read_options.column_names
                else:
                    raise ValueError("No column names found!")
            convert_options.column_types = {
                column_name: pa.string() for column_name in column_names
            }
            kwargs["convert_options"] = convert_options
        return kwargs


class ReadKwargsProviderPyArrowSchemaOverride(ContentTypeKwargsProvider):
    """ReadKwargsProvider impl that explicitly maps column names to column types when
    loading dataset files into a PyArrow table. Disables the default type inference
    behavior on the defined columns."""
    def __init__(self, schema: Optional[pa.Schema] = None):
        self.schema = schema

    def _get_kwargs(
            self,
            content_type: str,
            kwargs: Dict[str, Any]) -> Dict[str, Any]:
        if content_type in DELIMITED_TEXT_CONTENT_TYPES:
            convert_options = kwargs.get("convert_options", pacsv.ConvertOptions())
            if self.schema:
                convert_options.column_types = self.schema
            kwargs["convert_options"] = convert_options
        elif content_type == ContentType.PARQUET:
            # kwargs here are passed into `pyarrow.parquet.read_table`.
            # Only supported in PyArrow 8.0.0+
            if self.schema:
                kwargs["schema"] = self.schema
        return kwargs


def _add_column_kwargs(
        content_type: str,
        column_names: Optional[List[str]],
        include_columns: Optional[List[str]],
        kwargs: Dict[str, Any]):

    if content_type in DELIMITED_TEXT_CONTENT_TYPES:
        read_options: pacsv.ReadOptions = kwargs.get("read_options")
        if read_options is None:
            read_options = pacsv.ReadOptions()
        if column_names:
            read_options.column_names = column_names
        else:
            read_options.autogenerate_column_names = True
        kwargs["read_options"] = read_options
        convert_options: pacsv.ConvertOptions = kwargs.get("convert_options")
        if convert_options is None:
            convert_options = pacsv.ConvertOptions()
        if include_columns:
            convert_options.include_columns = include_columns
        kwargs["convert_options"] = convert_options
    else:
        if content_type in TABULAR_CONTENT_TYPES:
            kwargs["columns"] = include_columns
        else:
            if include_columns:
                logger.warning(
                    f"Ignoring request to include columns {include_columns} "
                    f"for non-tabular content type {content_type}")


def s3_file_to_table(
        s3_url: str,
        content_type: str,
        content_encoding: str,
        column_names: Optional[List[str]] = None,
        include_columns: Optional[List[str]] = None,
        pa_read_func_kwargs_provider: Optional[ReadKwargsProvider] = None,
        **s3_client_kwargs) -> pa.Table:

    from deltacat.aws import s3u as s3_utils
    logger.debug(f"Reading {s3_url} to PyArrow. Content type: {content_type}. "
                 f"Encoding: {content_encoding}")
    s3_obj = s3_utils.get_object_at_url(
        s3_url,
        **s3_client_kwargs
    )
    logger.debug(f"Read S3 object from {s3_url}: {s3_obj}")
    pa_read_func = CONTENT_TYPE_TO_PA_READ_FUNC[content_type]
    input_file_init = ENCODING_TO_FILE_INIT[content_encoding]
    input_file = input_file_init(fileobj=io.BytesIO(s3_obj["Body"].read()))

    args = [input_file]
    kwargs = content_type_to_reader_kwargs(content_type)
    _add_column_kwargs(content_type, column_names, include_columns, kwargs)

    if pa_read_func_kwargs_provider:
        kwargs = pa_read_func_kwargs_provider(content_type, kwargs)

    logger.debug(f"Reading {s3_url} via {pa_read_func} with kwargs: {kwargs}")
    table, latency = timed_invocation(
        pa_read_func,
        *args,
        **kwargs
    )
    logger.debug(f"Time to read {s3_url} into PyArrow table: {latency}s")
    return table


def table_size(table: pa.Table) -> int:
    return table.nbytes


def table_to_file(
        table: pa.Table,
        base_path: str,
        file_system: AbstractFileSystem,
        block_path_provider: BlockWritePathProvider,
        content_type: str = ContentType.PARQUET.value,
        **kwargs) -> None:
    """
    Writes the given Pyarrow Table to a file.
    """
    writer = CONTENT_TYPE_TO_PA_WRITE_FUNC.get(content_type)
    if not writer:
        raise NotImplementedError(
            f"Pyarrow writer for content type '{content_type}' not "
            f"implemented. Known content types: "
            f"{CONTENT_TYPE_TO_PA_WRITE_FUNC.keys}")
    path = block_path_provider(base_path)
    writer(
        table,
        path,
        filesystem=file_system,
        **kwargs
    )



import boto3
import logging
from boto3.resources.base import ServiceResource
from botocore.client import BaseClient
from botocore.config import Config
from boto3.exceptions import ResourceNotExistsError
from functools import lru_cache
from deltacat import logs
from deltacat.aws.constants import BOTO_MAX_RETRIES

from typing import Optional

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def _resource(
        name: str,
        region: Optional[str],
        **kwargs) -> ServiceResource:
    boto_config = Config(
        retries={
            "max_attempts": BOTO_MAX_RETRIES,
            "mode": 'standard'
        }
    )
    return boto3.resource(
        name,
        region,
        config=boto_config,
        **kwargs,
    )


def _client(
        name: str,
        region: Optional[str],
        **kwargs) -> BaseClient:
    try:
        # try to re-use a client from the resource cache first
        return resource_cache(name, region, **kwargs).meta.client
    except ResourceNotExistsError:
        # fall back for clients without an associated resource
        boto_config = Config(
            retries={
                "max_attempts": BOTO_MAX_RETRIES,
                "mode": 'standard'
            }
        )
        return boto3.client(
            name,
            region,
            config=boto_config,
            **kwargs,
        )


def resource_cache(
        name: str,
        region: Optional[str],
        **kwargs) -> ServiceResource:
    # we don't use the @lru_cache decorator because Ray can't pickle it
    cached_function = lru_cache()(_resource)
    return cached_function(name, region, **kwargs)


def client_cache(
        name: str,
        region: Optional[str],
        **kwargs) -> BaseClient:
    # we don't use the @lru_cache decorator because Ray can't pickle it
    cached_function = lru_cache()(_client)
    return cached_function(name, region, **kwargs)



import logging
import time
import functools
import ray

from collections import defaultdict

from deltacat import logs
from deltacat.compute.stats.models.delta_stats import DeltaStats
from deltacat.storage import Delta, DeltaLocator, Partition, \
    PartitionLocator, interface as unimplemented_deltacat_storage
from deltacat.utils.ray_utils.concurrency import invoke_parallel, \
    round_robin_options_provider
from deltacat.utils.ray_utils.runtime import live_node_resource_keys
from deltacat.compute.compactor.steps import hash_bucket as hb, dedupe as dd, \
    materialize as mat
from deltacat.compute.compactor import SortKey, PrimaryKeyIndexMeta, \
    PrimaryKeyIndexLocator, PrimaryKeyIndexVersionMeta, \
    PrimaryKeyIndexVersionLocator, RoundCompletionInfo, \
    PyArrowWriteResult
from deltacat.compute.compactor.utils import round_completion_file as rcf, io, \
    primary_key_index as pki
from deltacat.types.media import ContentType

from typing import List, Set, Optional, Tuple, Dict, Union, Any

import pyarrow as pa
logger = logs.configure_deltacat_logger(logging.getLogger(__name__))

_SORT_KEY_NAME_INDEX: int = 0
_SORT_KEY_ORDER_INDEX: int = 1
_PRIMARY_KEY_INDEX_ALGORITHM_VERSION: str = "1.0"


def check_preconditions(
        source_partition_locator: PartitionLocator,
        compacted_partition_locator: PartitionLocator,
        sort_keys: List[SortKey],
        max_records_per_output_file: int,
        new_hash_bucket_count: Optional[int],
        deltacat_storage=unimplemented_deltacat_storage) -> int:

    assert source_partition_locator.partition_values \
           == compacted_partition_locator.partition_values, \
        "In-place compaction must use the same partition values for the " \
        "source and destination."
    assert max_records_per_output_file >= 1, \
        "Max records per output file must be a positive value"
    if new_hash_bucket_count is not None:
        assert new_hash_bucket_count >= 1, \
            "New hash bucket count must be a positive value"
    return SortKey.validate_sort_keys(
        source_partition_locator,
        sort_keys,
        deltacat_storage,
    )


def compact_partition(
        source_partition_locator: PartitionLocator,
        compacted_partition_locator: PartitionLocator,
        primary_keys: Set[str],
        compaction_artifact_s3_bucket: str,
        last_stream_position_to_compact: int,
        hash_bucket_count: Optional[int] = None,
        sort_keys: List[SortKey] = None,
        records_per_primary_key_index_file: int = 38_000_000,
        records_per_compacted_file: int = 4_000_000,
        input_deltas_stats: Dict[int, DeltaStats] = None,
        min_pk_index_pa_bytes: int = 0,
        min_hash_bucket_chunk_size: int = 0,
        compacted_file_content_type: ContentType = ContentType.PARQUET,
        delete_prev_primary_key_index: bool = False,
        read_round_completion: bool = False,
        pg_config: Optional[List[Dict[str, Any]]] = None,
        schema_on_read: Optional[pa.schema] = None,  # TODO (ricmiyam): Remove this and retrieve schema from storage API
        deltacat_storage=unimplemented_deltacat_storage):

    logger.info(f"Starting compaction session for: {source_partition_locator}")
    partition = None
    compaction_rounds_executed = 0
    has_next_compaction_round = True
    opts={}
    if pg_config:
        opts=pg_config[0]
    while has_next_compaction_round:
        has_next_compaction_round_obj, new_partition_obj, new_rci_obj = \
            _execute_compaction_round.options(**opts).remote(
                source_partition_locator,
                compacted_partition_locator,
                primary_keys,
                compaction_artifact_s3_bucket,
                last_stream_position_to_compact,
                hash_bucket_count,
                sort_keys,
                records_per_primary_key_index_file,
                records_per_compacted_file,
                input_deltas_stats,
                min_pk_index_pa_bytes,
                min_hash_bucket_chunk_size,
                compacted_file_content_type,
                delete_prev_primary_key_index,
                read_round_completion,
                schema_on_read,
                deltacat_storage=deltacat_storage,
                pg_config=pg_config
            )
        has_next_compaction_round = ray.get(has_next_compaction_round_obj)
        new_partition = ray.get(new_partition_obj)
        new_rci = ray.get(new_rci_obj)
        if new_partition:
            partition = new_partition
            compacted_partition_locator = new_partition.locator
            compaction_rounds_executed += 1
        # Take new primary key index sizes into account for subsequent compaction rounds and their dedupe steps
        if new_rci:
            min_pk_index_pa_bytes = new_rci.pk_index_pyarrow_write_result.pyarrow_bytes

    logger.info(f"Partition-{source_partition_locator.partition_values}-> Compaction session data processing completed in "
                f"{compaction_rounds_executed} rounds.")
    if partition:
        logger.info(f"Committing compacted partition to: {partition.locator}")
        partition = deltacat_storage.commit_partition(partition)
        logger.info(f"Committed compacted partition: {partition}")
    logger.info(f"Completed compaction session for: {source_partition_locator}")

@ray.remote(num_cpus=0.1,num_returns=3)
def _execute_compaction_round(
        source_partition_locator: PartitionLocator,
        compacted_partition_locator: PartitionLocator,
        primary_keys: Set[str],
        compaction_artifact_s3_bucket: str,
        last_stream_position_to_compact: int,
        new_hash_bucket_count: Optional[int],
        sort_keys: List[SortKey],
        records_per_primary_key_index_file: int,
        records_per_compacted_file: int,
        input_deltas_stats: Dict[int, DeltaStats],
        min_pk_index_pa_bytes: int,
        min_hash_bucket_chunk_size: int,
        compacted_file_content_type: ContentType,
        delete_prev_primary_key_index: bool,
        read_round_completion: bool,
        schema_on_read: Optional[pa.schema],
        deltacat_storage = unimplemented_deltacat_storage,
        pg_config: Optional[List[Dict[str, Any]]] = None) \
        -> Tuple[bool, Optional[Partition], Optional[RoundCompletionInfo]]:


    if not primary_keys:
        # TODO (pdames): run simple rebatch to reduce all deltas into 1 delta
        #  with normalized manifest entry sizes
        raise NotImplementedError(
            "Compaction only supports tables with 1 or more primary keys")
    if sort_keys is None:
        sort_keys = []
    # TODO (pdames): detect and handle schema evolution (at least ensure that
    #  we don't recompact simple backwards-compatible changes like constraint
    #  widening and null column additions).
    # TODO (pdames): detect and optimize in-place compaction

    # check preconditions before doing any computationally expensive work
    bit_width_of_sort_keys = check_preconditions(
        source_partition_locator,
        compacted_partition_locator,
        sort_keys,
        records_per_compacted_file,
        new_hash_bucket_count,
        deltacat_storage,
    )

    # sort primary keys to produce the same pk digest regardless of input order
    primary_keys = sorted(primary_keys)

    # collect cluster resource stats
    # cluster_resources = ray.cluster_resources()
    # logger.info(f"Total cluster resources: {cluster_resources}")
    # logger.info(f"Available cluster resources: {ray.available_resources()}")
    # cluster_cpus = int(cluster_resources["CPU"])
    # logger.info(f"Total cluster CPUs: {cluster_cpus}")

    # collect node group resources

    cluster_resources = ray.cluster_resources()
    logger.info(f"Total cluster resources: {cluster_resources}")
    if pg_config: # use resource in each placement group
        node_resource_keys=None
        cluster_resources = pg_config[1]
        cluster_cpus = cluster_resources['CPU']   
    else: # use all cluster resource
        logger.info(f"Available cluster resources: {ray.available_resources()}")
        cluster_cpus = int(cluster_resources["CPU"])
        logger.info(f"Total cluster CPUs: {cluster_cpus}")
        node_resource_keys = live_node_resource_keys()
        logger.info(f"Found {len(node_resource_keys)} live cluster nodes: "
                   f"{node_resource_keys}") 

    if node_resource_keys:
        # create a remote options provider to round-robin tasks across all nodes
        logger.info(f"Setting round robin scheduling with node id:{node_resource_keys}")
        round_robin_opt_provider = functools.partial(
            round_robin_options_provider,
            resource_keys=node_resource_keys,
        )
    else:
        logger.info("Setting round robin scheduling to None")
        round_robin_opt_provider = None
    # assign a distinct index to each node in the cluster
    # head_node_ip = urllib.request.urlopen(
    #     "http://169.254.169.254/latest/meta-data/local-ipv4"
    # ).read().decode("utf-8")
    # print(f"head node ip: {head_node_ip}")

    # set max task parallelism equal to total cluster CPUs...
    # we assume here that we're running on a fixed-size cluster - this
    # assumption could be removed but we'd still need to know the maximum
    # "safe" number of parallel tasks that our autoscaling cluster could handle
    max_parallelism = int(cluster_cpus)
    logger.info(f"Max parallelism: {max_parallelism}")

    # get the root path of a compatible primary key index for this round
    compatible_primary_key_index_meta = PrimaryKeyIndexMeta.of(
        compacted_partition_locator,
        primary_keys,
        sort_keys,
        _PRIMARY_KEY_INDEX_ALGORITHM_VERSION,
    )
    compatible_primary_key_index_locator = PrimaryKeyIndexLocator.of(
        compatible_primary_key_index_meta)
    compatible_primary_key_index_root_path = \
        compatible_primary_key_index_locator.primary_key_index_root_path

    # read the results from any previously completed compaction round that used
    # a compatible primary key index
    round_completion_info = None
    if read_round_completion:
        round_completion_info = rcf.read_round_completion_file(
            compaction_artifact_s3_bucket,
            source_partition_locator,
            compatible_primary_key_index_root_path,
        )

    # read the previous compaction round's hash bucket count, if any
    old_hash_bucket_count = None
    if round_completion_info:
        old_pki_version_locator = round_completion_info\
            .primary_key_index_version_locator
        old_hash_bucket_count = old_pki_version_locator\
            .primary_key_index_version_meta \
            .hash_bucket_count
        min_pk_index_pa_bytes = round_completion_info.pk_index_pyarrow_write_result.pyarrow_bytes

    # use the new hash bucket count if provided, or fall back to old count
    hash_bucket_count = new_hash_bucket_count \
        if new_hash_bucket_count is not None \
        else old_hash_bucket_count

    # discover input delta files
    high_watermark = round_completion_info.high_watermark \
        if round_completion_info else None

    input_deltas = io.discover_deltas(
        source_partition_locator,
        high_watermark,
        last_stream_position_to_compact,
        deltacat_storage,
    )

    if not input_deltas:
        logger.info("No input deltas found to compact.")
        return False, None, None

    # limit the input deltas to fit on this cluster and convert them to
    # annotated deltas of equivalent size for easy parallel distribution

    uniform_deltas, hash_bucket_count, last_stream_position_compacted = \
        io.limit_input_deltas(
            input_deltas,
            cluster_resources,
            hash_bucket_count,
            min_pk_index_pa_bytes,
            min_hash_bucket_chunk_size,
            input_deltas_stats=input_deltas_stats,
            deltacat_storage=deltacat_storage
        )

    assert hash_bucket_count is not None and hash_bucket_count > 0, \
        f"Unexpected Error: Default hash bucket count ({hash_bucket_count}) " \
        f"is invalid."

    # rehash the primary key index if necessary
    round_completion_info = None
    if round_completion_info:
        logger.info(f"Round completion file contents: {round_completion_info}")
        # the previous primary key index is compatible with the current, but
        # will need to be rehashed if the hash bucket count has changed
        if hash_bucket_count != old_hash_bucket_count:
            round_completion_info = pki.rehash(
                round_robin_opt_provider,
                compaction_artifact_s3_bucket,
                source_partition_locator,
                round_completion_info,
                hash_bucket_count,
                max_parallelism,
                records_per_primary_key_index_file,
                delete_prev_primary_key_index,
            )
    else:
        logger.info(f"No prior round completion file found. Source partition: "
                    f"{source_partition_locator}. Primary key index locator: "
                    f"{compatible_primary_key_index_locator}")

    # parallel step 1:
    # group like primary keys together by hashing them into buckets
    hb_tasks_pending = invoke_parallel(
        items=uniform_deltas,
        ray_task=hb.hash_bucket,
        max_parallelism=max_parallelism,
        options_provider=round_robin_opt_provider,
        primary_keys=primary_keys,
        sort_keys=sort_keys,
        num_buckets=hash_bucket_count,
        num_groups=max_parallelism,
        deltacat_storage=deltacat_storage,
    )
    logger.info(f"Getting {len(hb_tasks_pending)} hash bucket results...")
    hb_results = ray.get([t[0] for t in hb_tasks_pending])
    logger.info(f"Got {len(hb_results)} hash bucket results.")
    all_hash_group_idx_to_obj_id = defaultdict(list)
    for hash_group_idx_to_obj_id in hb_results:
        for hash_group_index, object_id in enumerate(hash_group_idx_to_obj_id):
            if object_id:
                all_hash_group_idx_to_obj_id[hash_group_index].append(object_id)
    hash_group_count = dedupe_task_count = len(all_hash_group_idx_to_obj_id)
    logger.info(f"Hash bucket groups created: {hash_group_count}")

    # TODO (pdames): when resources are freed during the last round of hash
    #  bucketing, start running dedupe tasks that read existing dedupe
    #  output from S3 then wait for hash bucketing to finish before continuing

    # create a new stream for this round
    compacted_stream_locator = compacted_partition_locator.stream_locator
    stream = deltacat_storage.get_stream(
        compacted_stream_locator.namespace,
        compacted_stream_locator.table_name,
        compacted_stream_locator.table_version,
    )
    partition = deltacat_storage.stage_partition(
        stream,
        compacted_partition_locator.partition_values,
    )
    new_compacted_partition_locator = partition.locator

    # generate a new primary key index locator for this round
    new_primary_key_index_meta = PrimaryKeyIndexMeta.of(
        new_compacted_partition_locator,
        primary_keys,
        sort_keys,
        _PRIMARY_KEY_INDEX_ALGORITHM_VERSION,
    )
    new_primary_key_index_locator = PrimaryKeyIndexLocator.of(
        new_primary_key_index_meta)
    new_primary_key_index_root_path = new_primary_key_index_locator\
        .primary_key_index_root_path

    # generate a new primary key index version locator for this round
    new_primary_key_index_version_meta = PrimaryKeyIndexVersionMeta.of(
        new_primary_key_index_meta,
        hash_bucket_count,
    )
    new_pki_version_locator = PrimaryKeyIndexVersionLocator.generate(
        new_primary_key_index_version_meta)


    # parallel step 2:
    # discover records with duplicate primary keys in each hash bucket, and
    # identify the index of records to keep or drop based on sort keys
    num_materialize_buckets = max_parallelism
    logger.info(f"Materialize Bucket Count: {num_materialize_buckets}")
    record_counts_pending_materialize = \
        dd.RecordCountsPendingMaterialize.remote(dedupe_task_count)
    dd_tasks_pending = invoke_parallel(
        items=all_hash_group_idx_to_obj_id.values(),
        ray_task=dd.dedupe,
        max_parallelism=max_parallelism,
        options_provider=round_robin_opt_provider,
        kwargs_provider=lambda index, item: {"dedupe_task_index": index,
                                             "object_ids": item},
        compaction_artifact_s3_bucket=compaction_artifact_s3_bucket,
        round_completion_info=round_completion_info,
        new_primary_key_index_version_locator=new_pki_version_locator,
        sort_keys=sort_keys,
        max_records_per_index_file=records_per_primary_key_index_file,
        max_records_per_materialized_file=records_per_compacted_file,
        num_materialize_buckets=num_materialize_buckets,
        delete_old_primary_key_index=delete_prev_primary_key_index,
        record_counts_pending_materialize=record_counts_pending_materialize,
    )
    logger.info(f"Getting {len(dd_tasks_pending)} dedupe results...")
    dd_results = ray.get([t[0] for t in dd_tasks_pending])
    logger.info(f"Got {len(dd_results)} dedupe results.")
    all_mat_buckets_to_obj_id = defaultdict(list)
    for mat_bucket_idx_to_obj_id in dd_results:
        for bucket_idx, dd_task_index_and_object_id_tuple in \
                mat_bucket_idx_to_obj_id.items():
            all_mat_buckets_to_obj_id[bucket_idx].append(
                dd_task_index_and_object_id_tuple)
    logger.info(f"Getting {len(dd_tasks_pending)} dedupe result stat(s)...")
    pki_stats = ray.get([t[2] for t in dd_tasks_pending])
    logger.info(f"Got {len(pki_stats)} dedupe result stat(s).")
    logger.info(f"Materialize buckets created: "
                f"{len(all_mat_buckets_to_obj_id)}")

    # TODO(pdames): when resources are freed during the last round of deduping
    #  start running materialize tasks that read materialization source file
    #  tables from S3 then wait for deduping to finish before continuing

    # TODO(pdames): balance inputs to materialization tasks to ensure that each
    #  task has an approximately equal amount of input to materialize

    # TODO(pdames): garbage collect hash bucket output since it's no longer
    #  needed

    # parallel step 3:
    # materialize records to keep by index
    mat_tasks_pending = invoke_parallel(
        items=all_mat_buckets_to_obj_id.items(),
        ray_task=mat.materialize,
        max_parallelism=max_parallelism,
        options_provider=round_robin_opt_provider,
        kwargs_provider=lambda index, mat_bucket_idx_to_obj_id: {
            "mat_bucket_index": mat_bucket_idx_to_obj_id[0],
            "dedupe_task_idx_and_obj_id_tuples": mat_bucket_idx_to_obj_id[1],
        },
        schema=schema_on_read,
        round_completion_info=round_completion_info,
        source_partition_locator=source_partition_locator,
        partition=partition,
        max_records_per_output_file=records_per_compacted_file,
        compacted_file_content_type=compacted_file_content_type,
        deltacat_storage=deltacat_storage,
    )
    logger.info(f"Getting {len(mat_tasks_pending)} materialize result(s)...")
    mat_results = ray.get(mat_tasks_pending)
    logger.info(f"Got {len(mat_results)} materialize result(s).")

    mat_results = sorted(mat_results, key=lambda m: m.task_index)
    deltas = [m.delta for m in mat_results]
    merged_delta = Delta.merge_deltas(deltas)
    compacted_delta = deltacat_storage.commit_delta(merged_delta)
    logger.info(f"Committed compacted delta: {compacted_delta}")

    new_compacted_delta_locator = DeltaLocator.of(
        new_compacted_partition_locator,
        compacted_delta.stream_position,
    )

    round_completion_info = RoundCompletionInfo.of(
        last_stream_position_compacted,
        new_compacted_delta_locator,
        PyArrowWriteResult.union([m.pyarrow_write_result
                                  for m in mat_results]),
        PyArrowWriteResult.union(pki_stats),
        bit_width_of_sort_keys,
        new_pki_version_locator,
    )
    rcf.write_round_completion_file(
        compaction_artifact_s3_bucket,
        source_partition_locator,
        new_primary_key_index_root_path,
        round_completion_info,
    )
    time_mat_e = time.time()
    logger.info(f"partition-{source_partition_locator.partition_values},compacted at:{last_stream_position_compacted}, last position:{last_stream_position_to_compact}")
    return \
        (last_stream_position_compacted < last_stream_position_to_compact), \
        partition, \
        round_completion_info




# Allow self-referencing Type hints in Python 3.7.
from __future__ import annotations

from deltacat.storage import Table, TableVersion, Stream


class TableDefinition(dict):
    @staticmethod
    def of(table: Table,
           table_version: TableVersion,
           stream: Stream) -> TableDefinition:
        return TableDefinition({
            "table": table,
            "tableVersion": table_version,
            "stream": stream,
        })

    @property
    def table(self) -> Table:
        return self["table"]

    @property
    def table_version(self) -> TableVersion:
        return self["tableVersion"]

    @property
    def stream(self) -> Stream:
        return self["stream"]



import logging

import deltacat.logs
from deltacat.types.media import ContentType, ContentEncoding, TableType
from deltacat.types.tables import TableWriteMode
from deltacat.storage import ListResult, Namespace, LifecycleState, \
    SchemaConsistencyType, LocalTable, LocalDataset, DistributedDataset
from deltacat.compute.compactor import SortKey, SortOrder
from deltacat.catalog.model.table_definition import TableDefinition
from deltacat.catalog.model.catalog import Catalog, Catalogs, all_catalogs, \
    init
from deltacat.catalog.delegate import alter_table, create_table, drop_table, \
    refresh_table, list_tables, get_table, truncate_table, rename_table, \
    table_exists, list_namespaces, alter_namespace, create_namespace, \
    drop_namespace, default_namespace, get_namespace, namespace_exists, \
    write_to_table, read_table

logs.configure_deltacat_logger(logging.getLogger(__name__))

__version__ = "0.1.11"


__all__ = [
    "__version__",
    "all_catalogs",
    "alter_table",
    "create_table",
    "drop_table",
    "refresh_table",
    "list_tables",
    "get_table",
    "truncate_table",
    "rename_table",
    "table_exists",
    "list_namespaces",
    "get_namespace",
    "namespace_exists",
    "alter_namespace",
    "create_namespace",
    "drop_namespace",
    "default_namespace",
    "write_to_table",
    "read_table",
    "init",
    "Catalog",
    "ContentType",
    "ContentEncoding",
    "DistributedDataset",
    "LifecycleState",
    "ListResult",
    "LocalDataset",
    "LocalTable",
    "Namespace",
    "SchemaConsistencyType",
    "SortKey",
    "SortOrder",
    "TableDefinition",
    "TableType",
    "TableWriteMode",
]



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

import logging
from deltacat import logs
from deltacat.storage import DeltaType, Manifest, ManifestEntry, \
    ManifestEntryList
from typing import List, Optional, Callable, Union
from types import FunctionType
from deltacat.storage import Delta

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


class DeltaAnnotation(tuple):
    @staticmethod
    def of(file_index: int,
           delta_type: DeltaType,
           stream_position: int) -> DeltaAnnotation:
        return DeltaAnnotation((file_index, delta_type, stream_position))

    @property
    def annotation_file_index(self) -> int:
        return self[0]

    @property
    def annotation_delta_type(self) -> Optional[DeltaType]:
        return self[1]

    @property
    def annotation_stream_position(self) -> Optional[int]:
        return self[2]


class DeltaAnnotated(Delta):
    @staticmethod
    def of(delta: Delta) -> DeltaAnnotated:
        """
        Creates an annotated delta built from the input delta, which saves all
        properties for each delta manifest entry. All operations performed on
        the annotated delta by this module will preserve a mapping back to the
        original delta manifest entry indices and properties.
        """
        delta_annotated = DeltaAnnotated()
        delta_annotated.update(delta)
        entries = delta.manifest.entries
        _annotations = []
        if entries:
            dtype = delta.type
            pos = delta.stream_position
            _annotations = [DeltaAnnotation.of(i, dtype, pos) for i in
                            range(len(entries))]
        delta_annotated.annotations = _annotations
        return delta_annotated

    @staticmethod
    def rebatch(
            annotated_deltas: List[DeltaAnnotated],
            min_delta_bytes,
            min_file_counts: Optional[Union[int, float]]=float("inf"),
            estimation_function: Optional[Callable]=None) -> List[DeltaAnnotated]:
        """
        Simple greedy algorithm to split/merge 1 or more annotated deltas into
        size-limited annotated deltas. All ordered manifest entries in the input
        annotated deltas are appended to an annotated delta until
        delta_size_bytes >= min_delta_bytes, then a new delta is started. Note
        that byte size is measured in terms of manifest entry content length,
        which is expected to be equal to the number of bytes at rest for the
        associated object. Returns the list of annotated delta groups.
        """
        groups = []
        new_da = DeltaAnnotated()
        new_da_bytes = 0
        da_group_entry_count = 0
        for src_da in annotated_deltas:
            src_da_annotations = src_da.annotations
            src_da_entries = src_da.manifest.entries
            assert(len(src_da_annotations) == len(src_da_entries),
                   f"Unexpected Error: Length of delta annotations "
                   f"({len(src_da_annotations)}) doesn't mach the length of "
                   f"delta manifest entries ({len(src_da_entries)}).")
            for i, src_entry in enumerate(src_da_entries):
                DeltaAnnotated._append_annotated_entry(
                    src_da,
                    new_da,
                    src_entry,
                    src_da_annotations[i])
                # TODO: Fetch s3_obj["Size"] if entry content length undefined?
                estimated_new_da_bytes = estimation_function(src_entry.meta.content_length) if type(
                    estimation_function) is FunctionType else src_entry.meta.content_length
                new_da_bytes += estimated_new_da_bytes
                da_group_entry_count += 1
                if new_da_bytes >= min_delta_bytes or da_group_entry_count >= min_file_counts:
                    if new_da_bytes >= min_delta_bytes:
                        logger.info(
                            f"Appending group of {da_group_entry_count} elements "
                            f"and {new_da_bytes} bytes to meet file size limit")
                    if da_group_entry_count >= min_file_counts:
                        logger.info(
                            f"Appending group of {da_group_entry_count} elements "
                            f"and {da_group_entry_count} files to meet file count limit")
                    groups.append(new_da)
                    new_da = DeltaAnnotated()
                    new_da_bytes = 0
                    da_group_entry_count = 0
        if new_da:
            groups.append(new_da)
        return groups

    @staticmethod
    def split(
            src_da: DeltaAnnotated,
            pieces: int) -> List[DeltaAnnotated]:
        groups = []
        new_da = DeltaAnnotated()
        da_group_entry_count = 0
        src_da_annotations = src_da.annotations
        src_da_entries = src_da.manifest.entries
        assert (len(src_da_annotations) == len(src_da_entries),
                f"Unexpected Error: Length of delta annotations "
                f"({len(src_da_annotations)}) doesn't mach the length of "
                f"delta manifest entries ({len(src_da_entries)}).")
        src_da_entries_length = len(src_da_entries)
        equal_length = src_da_entries_length // pieces
        for i in range(len(src_da_entries)):
            DeltaAnnotated._append_annotated_entry(
                src_da,
                new_da,
                src_da_entries[i],
                src_da_annotations[i])
            # TODO: Fetch s3_obj["Size"] if entry content length undefined?
            da_group_entry_count += 1
            if da_group_entry_count >= equal_length and i < equal_length * (pieces - 1):
                logger.info(
                    f"Splitting {da_group_entry_count} manifest files "
                    f"to {pieces} pieces of {equal_length} size.")
                groups.append(new_da)
                new_da = DeltaAnnotated()
                da_group_entry_count = 0
            if i == len(src_da_entries) - 1:
                groups.append(new_da)
                logger.info(
                    f"Splitting {da_group_entry_count} manifest files "
                    f"to {pieces} pieces of {equal_length} size.")
                new_da = DeltaAnnotated()
        if new_da:
            groups.append(new_da)
        return groups

    @property
    def annotations(self) -> List[DeltaAnnotation]:
        return self["annotations"]

    @annotations.setter
    def annotations(self, _annotations: List[DeltaAnnotation]):
        self["annotations"] = _annotations

    @staticmethod
    def _append_annotated_entry(
            src_da: DeltaAnnotated,
            dst_da: DeltaAnnotated,
            src_entry: ManifestEntry,
            src_annotation: DeltaAnnotation):

        if not dst_da:
            # copy all extended properties from the source delta manifest impl
            dst_da.update(src_da)
            dst_da.manifest = Manifest.of(ManifestEntryList([src_entry]))
            dst_da.annotations = [src_annotation]
        else:
            entries = dst_da.manifest.entries
            src_dl = src_da.locator
            dst_dl = dst_da.locator
            # remove delta type and stream position if there is a conflict
            if src_da.type != dst_da.type:
                dst_da.type = None
            if src_dl.stream_position != dst_dl.stream_position:
                dst_dl.stream_position = None
            entries.append(src_entry)
            dst_da.annotations.append(src_annotation)



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

import pyarrow as pa

from deltacat.storage.model.namespace import NamespaceLocator
from deltacat.storage.model.stream import StreamLocator
from deltacat.storage.model.table import TableLocator
from deltacat.storage.model.table_version import TableVersionLocator
from deltacat.storage.model.types import CommitState
from deltacat.storage.model.locator import Locator
from deltacat.types.media import ContentType

from typing import Any, Dict, List, Optional, Union


class Partition(dict):
    @staticmethod
    def of(locator: Optional[PartitionLocator],
           schema: Optional[Union[pa.Schema, str, bytes]],
           content_types: Optional[List[ContentType]],
           state: Optional[CommitState] = None,
           previous_stream_position: Optional[int] = None,
           previous_partition_id: Optional[str] = None,
           stream_position: Optional[int] = None,
           next_partition_id: Optional[str] = None) -> Partition:
        partition = Partition()
        partition.locator = locator
        partition.schema = schema
        partition.content_types = content_types
        partition.state = state
        partition.previous_stream_position = previous_stream_position
        partition.previous_partition_id = previous_partition_id
        partition.stream_position = stream_position
        partition.next_partition_id = next_partition_id
        return partition

    @property
    def locator(self) -> Optional[PartitionLocator]:
        val: Dict[str, Any] = self.get("partitionLocator")
        if val is not None and not isinstance(val, PartitionLocator):
            self.locator = val = PartitionLocator(val)
        return val

    @locator.setter
    def locator(
            self,
            partition_locator: Optional[PartitionLocator]) -> None:
        self["partitionLocator"] = partition_locator

    @property
    def schema(self) -> Optional[Union[pa.Schema, str, bytes]]:
        return self.get("schema")

    @schema.setter
    def schema(self, schema: Optional[Union[pa.Schema, str, bytes]]) -> None:
        self["schema"] = schema

    @property
    def content_types(self) -> Optional[List[ContentType]]:
        content_types = self.get("contentTypes")
        return None if content_types is None else \
            [None if _ is None else ContentType(_) for _ in content_types]

    @content_types.setter
    def content_types(
            self,
            supported_content_types: Optional[List[ContentType]]) -> None:
        self["contentTypes"] = supported_content_types

    @property
    def state(self) -> Optional[CommitState]:
        state = self.get("state")
        return None if state is None else CommitState(state)

    @state.setter
    def state(self, state: Optional[CommitState]) -> None:
        self["state"] = state

    @property
    def previous_stream_position(self) -> Optional[int]:
        return self.get("previousStreamPosition")

    @previous_stream_position.setter
    def previous_stream_position(
            self,
            previous_stream_position: Optional[int]) -> None:
        self["previousStreamPosition"] = previous_stream_position

    @property
    def previous_partition_id(self) -> Optional[str]:
        return self.get("previousPartitionId")

    @previous_partition_id.setter
    def previous_partition_id(
            self,
            previous_partition_id: Optional[str]) -> None:
        self["previousPartitionId"] = previous_partition_id

    @property
    def stream_position(self) -> Optional[int]:
        return self.get("streamPosition")

    @stream_position.setter
    def stream_position(self, stream_position: Optional[int]):
        self["streamPosition"] = stream_position

    @property
    def next_partition_id(self) -> Optional[str]:
        return self.get("nextPartitionId")

    @next_partition_id.setter
    def next_partition_id(self, next_partition_id: Optional[str]):
        self["nextPartitionId"] = next_partition_id

    @property
    def partition_id(self) -> Optional[str]:
        partition_locator = self.locator
        if partition_locator:
            return partition_locator.partition_id
        return None

    @property
    def stream_id(self) -> Optional[str]:
        partition_locator = self.locator
        if partition_locator:
            return partition_locator.stream_id
        return None

    @property
    def partition_values(self) -> Optional[List[Any]]:
        partition_locator = self.locator
        if partition_locator:
            return partition_locator.partition_values

    @property
    def namespace_locator(self) -> Optional[NamespaceLocator]:
        partition_locator = self.locator
        if partition_locator:
            return partition_locator.namespace_locator
        return None

    @property
    def table_locator(self) -> Optional[TableLocator]:
        partition_locator = self.locator
        if partition_locator:
            return partition_locator.table_locator
        return None

    @property
    def table_version_locator(self) -> Optional[TableVersionLocator]:
        partition_locator = self.locator
        if partition_locator:
            return partition_locator.table_version_locator
        return None

    @property
    def stream_locator(self) -> Optional[StreamLocator]:
        partition_locator = self.locator
        if partition_locator:
            return partition_locator.stream_locator
        return None

    @property
    def storage_type(self) -> Optional[str]:
        partition_locator = self.locator
        if partition_locator:
            return partition_locator.storage_type
        return None

    @property
    def namespace(self) -> Optional[str]:
        partition_locator = self.locator
        if partition_locator:
            return partition_locator.namespace
        return None

    @property
    def table_name(self) -> Optional[str]:
        partition_locator = self.locator
        if partition_locator:
            return partition_locator.table_name
        return None

    @property
    def table_version(self) -> Optional[str]:
        partition_locator = self.locator
        if partition_locator:
            return partition_locator.table_version
        return None

    def is_supported_content_type(self, content_type: ContentType) -> bool:
        supported_content_types = self.content_types
        return (not supported_content_types) or \
               (content_type in supported_content_types)


class PartitionLocator(Locator, dict):
    @staticmethod
    def of(stream_locator: Optional[StreamLocator],
           partition_values: Optional[List[Any]],
           partition_id: Optional[str]) -> PartitionLocator:
        """
        Creates a stream partition locator. Partition ID is
        case-sensitive.

        Partition Value types must ensure that
        `str(partition_value1) == str(partition_value2)` always implies
        `partition_value1 == partition_value2` (i.e. if two string
        representations of partition values are equal, then the two partition
        values are equal).
        """
        partition_locator = PartitionLocator()
        partition_locator.stream_locator = stream_locator
        partition_locator.partition_values = partition_values
        partition_locator.partition_id = partition_id
        return partition_locator

    @staticmethod
    def at(namespace: Optional[str],
           table_name: Optional[str],
           table_version: Optional[str],
           stream_id: Optional[str],
           storage_type: Optional[str],
           partition_values: Optional[List[Any]],
           partition_id: Optional[str]) -> PartitionLocator:
        stream_locator = StreamLocator.at(
            namespace,
            table_name,
            table_version,
            stream_id,
            storage_type,
        )
        return PartitionLocator.of(
            stream_locator,
            partition_values,
            partition_id,
        )

    @property
    def stream_locator(self) -> Optional[StreamLocator]:
        val: Dict[str, Any] = self.get("streamLocator")
        if val is not None and not isinstance(val, StreamLocator):
            self.stream_locator = val = StreamLocator(val)
        return val

    @stream_locator.setter
    def stream_locator(
            self,
            stream_locator: Optional[StreamLocator]) -> None:
        self["streamLocator"] = stream_locator

    @property
    def partition_values(self) -> Optional[List[Any]]:
        return self.get("partitionValues")

    @partition_values.setter
    def partition_values(self, partition_values: Optional[List[Any]]) -> None:
        self["partitionValues"] = partition_values

    @property
    def partition_id(self) -> Optional[str]:
        return self.get("partitionId")

    @partition_id.setter
    def partition_id(
            self,
            partition_id: Optional[str]) -> None:
        self["partitionId"] = partition_id

    @property
    def namespace_locator(self) -> Optional[NamespaceLocator]:
        stream_locator = self.stream_locator
        if stream_locator:
            return stream_locator.namespace_locator
        return None

    @property
    def table_locator(self) -> Optional[TableLocator]:
        stream_locator = self.stream_locator
        if stream_locator:
            return stream_locator.table_locator
        return None

    @property
    def table_version_locator(self) -> Optional[TableVersionLocator]:
        stream_locator = self.stream_locator
        if stream_locator:
            return stream_locator.table_version_locator
        return None

    @property
    def stream_id(self) -> Optional[str]:
        stream_locator = self.stream_locator
        if stream_locator:
            return stream_locator.stream_id
        return None

    @property
    def storage_type(self) -> Optional[str]:
        stream_locator = self.stream_locator
        if stream_locator:
            return stream_locator.storage_type
        return None

    @property
    def namespace(self) -> Optional[str]:
        stream_locator = self.stream_locator
        if stream_locator:
            return stream_locator.namespace
        return None

    @property
    def table_name(self) -> Optional[str]:
        stream_locator = self.stream_locator
        if stream_locator:
            return stream_locator.table_name
        return None

    @property
    def table_version(self) -> Optional[str]:
        stream_locator = self.stream_locator
        if stream_locator:
            return stream_locator.table_version
        return None

    def canonical_string(self) -> str:
        """
        Returns a unique string for the given locator that can be used
        for equality checks (i.e. two locators are equal if they have
        the same canonical string).
        """
        sl_hexdigest = self.stream_locator.hexdigest()
        partition_vals = str(self.partition_values)
        partition_id = self.partition_id
        return f"{sl_hexdigest}|{partition_vals}|{partition_id}"



import ray
import deltacat.aws.clients as aws_utils
import logging
import multiprocessing
import s3fs
import pyarrow as pa

from functools import partial
from uuid import uuid4

from ray.types import ObjectRef
from ray.data.datasource import BlockWritePathProvider
from ray.data.block import Block, BlockAccessor, BlockMetadata

from deltacat import logs
from deltacat.storage import LocalTable, LocalDataset, DistributedDataset, \
    Manifest, ManifestEntry, ManifestEntryList
from deltacat.aws.constants import TIMEOUT_ERROR_CODES
from deltacat.exceptions import RetryableError, NonRetryableError
from deltacat.types.media import ContentType, ContentEncoding
from deltacat.types.tables import TABLE_TYPE_TO_READER_FUNC, \
    TABLE_CLASS_TO_SIZE_FUNC, get_table_length
from deltacat.types.media import TableType
from deltacat.utils.common import ReadKwargsProvider

from boto3.resources.base import ServiceResource
from botocore.client import BaseClient
from botocore.exceptions import ClientError
from tenacity import Retrying
from tenacity import wait_random_exponential
from tenacity import stop_after_delay
from tenacity import retry_if_exception_type, retry_if_not_exception_type

from typing import Any, Callable, Dict, List, Optional, Generator, Union

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


@ray.remote
class CapturedBlockWritePaths:
    def __init__(self):
        self._write_paths: List[str] = []
        self._block_refs: List[ObjectRef[Block]] = []

    def extend(
            self,
            write_paths: List[str],
            block_refs: List[ObjectRef[Block]]):
        try:
            iter(write_paths)
        except TypeError:
            pass
        else:
            self._write_paths.extend(write_paths)
        try:
            iter(block_refs)
        except TypeError:
            pass
        else:
            self._block_refs.extend(block_refs)

    def write_paths(self) -> List[str]:
        return self._write_paths

    def block_refs(self) -> List[ObjectRef[Block]]:
        return self._block_refs


class UuidBlockWritePathProvider(BlockWritePathProvider):
    """Block write path provider implementation that writes each
    dataset block out to a file of the form: {base_path}/{uuid}
    """
    def __init__(self, capture_actor: CapturedBlockWritePaths):
        self.write_paths: List[str] = []
        self.block_refs: List[ObjectRef[Block]] = []
        self.capture_actor = capture_actor

    def __del__(self):
        if self.write_paths or self.block_refs:
            self.capture_actor.extend.remote(
                self.write_paths,
                self.block_refs,
            )

    def _get_write_path_for_block(
            self,
            base_path: str,
            *,
            filesystem: Optional[pa.filesystem.FileSystem] = None,
            dataset_uuid: Optional[str] = None,
            block: Optional[ObjectRef[Block]] = None,
            block_index: Optional[int] = None,
            file_format: Optional[str] = None) -> str:
        write_path = f"{base_path}/{str(uuid4())}"
        self.write_paths.append(write_path)
        if block:
            self.block_refs.append(block)
        return write_path


class S3Url:
    def __init__(
            self,
            url: str):

        from urllib.parse import urlparse

        self._parsed = urlparse(
            url,
            allow_fragments=False  # support '#' in path
        )
        if not self._parsed.scheme:  # support paths w/o 's3://' scheme
            url = f"s3://{url}"
            self._parsed = urlparse(url, allow_fragments=False)
        if self._parsed.query:  # support '?' in path
            self.key = \
                f"{self._parsed.path.lstrip('/')}?{self._parsed.query}"
        else:
            self.key = self._parsed.path.lstrip('/')
        self.bucket = self._parsed.netloc
        self.url = self._parsed.geturl()


def parse_s3_url(url: str) -> S3Url:
    return S3Url(url)


def s3_resource_cache(
        region: Optional[str],
        **kwargs) -> ServiceResource:

    return aws_utils.resource_cache(
        "s3",
        region,
        **kwargs,
    )


def s3_client_cache(
        region: Optional[str],
        **kwargs) -> BaseClient:

    return aws_utils.client_cache(
        "s3",
        region,
        **kwargs
    )


def get_object_at_url(
        url: str,
        **s3_client_kwargs) -> Dict[str, Any]:

    s3 = s3_client_cache(
        None,
        **s3_client_kwargs)

    parsed_s3_url = parse_s3_url(url)
    return s3.get_object(
        Bucket=parsed_s3_url.bucket,
        Key=parsed_s3_url.key
    )


def delete_files_by_prefix(
        bucket: str,
        prefix: str,
        **s3_client_kwargs) -> None:

    s3 = s3_resource_cache(None, **s3_client_kwargs)
    bucket = s3.Bucket(bucket)
    bucket.objects.filter(Prefix=prefix).delete()


def filter_paths_by_prefix(bucket, prefix):
    return objects_to_paths(
        bucket,
        filter_objects_by_prefix(bucket, prefix),
    )


def objects_to_paths(bucket, objects):
    for obj in objects:
        yield get_path_from_object(bucket, obj)


def get_path_from_object(bucket, obj):
    return "s3://{}/{}".format(bucket, obj["Key"])


def filter_objects_by_prefix(
        bucket: str,
        prefix: str,
        **s3_client_kwargs) -> Generator[Dict[str, Any], None, None]:

    s3 = s3_client_cache(
        None,
        **s3_client_kwargs
    )
    params = {"Bucket": bucket, "Prefix": prefix}
    more_objects_to_list = True
    while more_objects_to_list:
        response = s3.list_objects_v2(**params)
        if "Contents" in response:
            for obj in response["Contents"]:
                yield obj
        params["ContinuationToken"] = response.get("NextContinuationToken")
        more_objects_to_list = params["ContinuationToken"] is not None


def read_file(
        s3_url: str,
        content_type: ContentType,
        content_encoding: ContentEncoding = ContentEncoding.IDENTITY,
        table_type: TableType = TableType.PYARROW,
        column_names: Optional[List[str]] = None,
        include_columns: Optional[List[str]] = None,
        file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None,
        **s3_client_kwargs) -> LocalTable:

    reader = TABLE_TYPE_TO_READER_FUNC[table_type.value]
    try:
        table = reader(
            s3_url,
            content_type.value,
            content_encoding.value,
            column_names,
            include_columns,
            file_reader_kwargs_provider,
            **s3_client_kwargs
        )
        return table
    except ClientError as e:
        if e.response["Error"]["Code"] in TIMEOUT_ERROR_CODES:
            # Timeout error not caught by botocore
            raise RetryableError(f"Retry table download from: {s3_url}") \
                from e
        raise NonRetryableError(f"Failed table download from: {s3_url}") \
            from e


def upload_sliced_table(
        table: Union[LocalTable, DistributedDataset],
        s3_url_prefix: str,
        s3_file_system: s3fs.S3FileSystem,
        max_records_per_entry: Optional[int],
        s3_table_writer_func: Callable,
        table_slicer_func: Callable,
        s3_table_writer_kwargs: Optional[Dict[str, Any]] = None,
        content_type: ContentType = ContentType.PARQUET,
        **s3_client_kwargs) -> ManifestEntryList:

    # @retry decorator can't be pickled by Ray, so wrap upload in Retrying
    retrying = Retrying(
        wait=wait_random_exponential(multiplier=1, max=60),
        stop=stop_after_delay(30 * 60),
        retry=retry_if_exception_type(RetryableError)
    )

    manifest_entries = ManifestEntryList()
    table_record_count = get_table_length(table)

    if max_records_per_entry is None or not table_record_count:
        # write the whole table to a single s3 file
        manifest_entries = retrying(
            upload_table,
            table,
            f"{s3_url_prefix}",
            s3_file_system,
            s3_table_writer_func,
            s3_table_writer_kwargs,
            content_type,
            **s3_client_kwargs
        )
    else:
        # iteratively write table slices
        table_slices = table_slicer_func(
            table,
            max_records_per_entry
        )
        for table_slice in table_slices:
            slice_entries = retrying(
                upload_table,
                table_slice,
                f"{s3_url_prefix}",
                s3_file_system,
                s3_table_writer_func,
                s3_table_writer_kwargs,
                content_type,
                **s3_client_kwargs
            )
            manifest_entries.extend(slice_entries)

    return manifest_entries


@ray.remote
def _block_metadata(block: Block) -> BlockMetadata:
    return BlockAccessor.for_block(block).get_metadata(
        input_files=None,
        exec_stats=None,
    )


def _get_metadata(
        table: Union[LocalTable, DistributedDataset],
        write_paths: List[str],
        block_refs: List[ObjectRef[Block]])-> List[BlockMetadata]:
    metadata: List[BlockMetadata] = []
    if not block_refs:
        # this must be a local table - ensure it was written to only 1 file
        assert len(write_paths) == 1, \
            f"Expected table of type '{type(table)}' to be written to 1 " \
            f"file, but found {len(write_paths)} files."
        table_size = None
        table_size_func = TABLE_CLASS_TO_SIZE_FUNC.get(type(table))
        if table_size_func:
            table_size = table_size_func(table)
        else:
            logger.warning(f"Unable to estimate '{type(table)}' table size.")
        metadata.append(
            BlockMetadata(
                num_rows=get_table_length(table),
                size_bytes=table_size,
                schema=None,
                input_files=None,
                exec_stats=None,
            )
        )
    else:
        # TODO(pdames): Expose BlockList metadata getter from Ray Dataset?
        # ray 1.10
        # metadata = dataset._blocks.get_metadata()
        # ray 2.0.0dev
        metadata = table._plan.execute().get_metadata()
        if not metadata or metadata[0].size_bytes is None or \
                metadata[0].num_rows is None:
            metadata_futures = [_block_metadata.remote(block_ref)
                                for block_ref
                                in block_refs]
            metadata = ray.get(metadata_futures)
    return metadata


def upload_table(
        table: Union[LocalTable, DistributedDataset],
        s3_base_url: str,
        s3_file_system: s3fs.S3FileSystem,
        s3_table_writer_func: Callable,
        s3_table_writer_kwargs: Optional[Dict[str, Any]],
        content_type: ContentType = ContentType.PARQUET,
        **s3_client_kwargs) -> ManifestEntryList:
    """
    Writes the given table to 1 or more S3 files and return Redshift
    manifest entries describing the uploaded files.
    """
    if s3_table_writer_kwargs is None:
        s3_table_writer_kwargs = {}

    capture_actor = CapturedBlockWritePaths.remote()
    block_write_path_provider = UuidBlockWritePathProvider(capture_actor)
    s3_table_writer_func(
        table,
        s3_base_url,
        s3_file_system,
        block_write_path_provider,
        content_type.value,
        **s3_table_writer_kwargs
    )
    # TODO: Add a proper fix for block_refs and write_paths not persisting in Ray actors
    del block_write_path_provider
    block_refs = ray.get(capture_actor.block_refs.remote())
    write_paths = ray.get(capture_actor.write_paths.remote())
    metadata = _get_metadata(table, write_paths, block_refs)
    manifest_entries = ManifestEntryList()
    for block_idx, s3_url in enumerate(write_paths):
        try:
            manifest_entry = ManifestEntry.from_s3_obj_url(
                s3_url,
                metadata[block_idx].num_rows,
                metadata[block_idx].size_bytes,
                **s3_client_kwargs,
            )
            manifest_entries.append(manifest_entry)
        except ClientError as e:
            if e.response["Error"]["Code"] == "NoSuchKey":
                # s3fs may swallow S3 errors - we were probably throttled
                raise RetryableError(f"Retry table upload to: {s3_url}") \
                    from e
            raise NonRetryableError(f"Failed table upload to: {s3_url}") \
                from e
    return manifest_entries


def download_manifest_entry(
        manifest_entry: ManifestEntry,
        token_holder: Optional[Dict[str, Any]] = None,
        table_type: TableType = TableType.PYARROW,
        column_names: Optional[List[str]] = None,
        include_columns: Optional[List[str]] = None,
        file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None,
        content_type: Optional[ContentType] = None,
        content_encoding: Optional[ContentEncoding] = None) -> LocalTable:

    s3_client_kwargs = {
        "aws_access_key_id": token_holder["accessKeyId"],
        "aws_secret_access_key": token_holder["secretAccessKey"],
        "aws_session_token": token_holder["sessionToken"]
    } if token_holder else {}
    if not content_type:
        content_type = manifest_entry.meta.content_type
        assert content_type, \
            f"Unknown content type for manifest entry: {manifest_entry}"
        content_type = ContentType(content_type)
    if not content_encoding:
        content_encoding = manifest_entry.meta.content_encoding
        assert content_encoding, \
            f"Unknown content encoding for manifest entry: {manifest_entry}"
        content_encoding = ContentEncoding(content_encoding)
    s3_url = manifest_entry.uri
    if s3_url is None:
        s3_url = manifest_entry.url
    # @retry decorator can't be pickled by Ray, so wrap download in Retrying
    retrying = Retrying(
        wait=wait_random_exponential(multiplier=1, max=60),
        stop=stop_after_delay(30 * 60),
        retry=retry_if_not_exception_type(NonRetryableError)
    )
    table = retrying(
        read_file,
        s3_url,
        content_type,
        content_encoding,
        table_type,
        column_names,
        include_columns,
        file_reader_kwargs_provider,
        **s3_client_kwargs,
    )
    return table


def _download_manifest_entries(
        manifest: Manifest,
        token_holder: Optional[Dict[str, Any]] = None,
        table_type: TableType = TableType.PYARROW,
        column_names: Optional[List[str]] = None,
        include_columns: Optional[List[str]] = None,
        file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None) \
        -> LocalDataset:

    return [
        download_manifest_entry(e, token_holder, table_type, column_names,
                                include_columns, file_reader_kwargs_provider)
        for e in manifest.entries
    ]


def _download_manifest_entries_parallel(
        manifest: Manifest,
        token_holder: Optional[Dict[str, Any]] = None,
        table_type: TableType = TableType.PYARROW,
        max_parallelism: Optional[int] = None,
        column_names: Optional[List[str]] = None,
        include_columns: Optional[List[str]] = None,
        file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None) \
        -> LocalDataset:

    tables = []
    pool = multiprocessing.Pool(max_parallelism)
    downloader = partial(
        download_manifest_entry,
        token_holder=token_holder,
        table_type=table_type,
        column_names=column_names,
        include_columns=include_columns,
        file_reader_kwargs_provider=file_reader_kwargs_provider,
    )
    for table in pool.map(downloader, [e for e in manifest.entries]):
        tables.append(table)
    return tables


def download_manifest_entries(
        manifest: Manifest,
        token_holder: Optional[Dict[str, Any]] = None,
        table_type: TableType = TableType.PYARROW,
        max_parallelism: Optional[int] = 1,
        column_names: Optional[List[str]] = None,
        include_columns: Optional[List[str]] = None,
        file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None) \
        -> LocalDataset:

    if max_parallelism and max_parallelism <= 1:
        return _download_manifest_entries(
            manifest,
            token_holder,
            table_type,
            column_names,
            include_columns,
            file_reader_kwargs_provider,
        )
    else:
        return _download_manifest_entries_parallel(
            manifest,
            token_holder,
            table_type,
            max_parallelism,
            column_names,
            include_columns,
            file_reader_kwargs_provider,
        )


def upload(
        s3_url: str,
        body,
        **s3_client_kwargs) -> Dict[str, Any]:

    # TODO (pdames): add tenacity retrying
    parsed_s3_url = parse_s3_url(s3_url)
    s3 = s3_client_cache(None, **s3_client_kwargs)
    return s3.put_object(
        Body=body,
        Bucket=parsed_s3_url.bucket,
        Key=parsed_s3_url.key,
    )


def download(
        s3_url: str,
        fail_if_not_found: bool = True,
        **s3_client_kwargs) -> Optional[Dict[str, Any]]:

    # TODO (pdames): add tenacity retrying
    parsed_s3_url = parse_s3_url(s3_url)
    s3 = s3_client_cache(None, **s3_client_kwargs)
    try:
        return s3.get_object(
            Bucket=parsed_s3_url.bucket,
            Key=parsed_s3_url.key,
        )
    except ClientError as e:
        if fail_if_not_found:
            raise
        else:
            if e.response['Error']['Code'] != "404":
                if e.response['Error']['Code'] != 'NoSuchKey':
                    raise
            logger.info(
                f"file not found: {s3_url}")
    except s3.exceptions.NoSuchKey:
        if fail_if_not_found:
            raise
        else:
            logger.info(
                f"file not found: {s3_url}")
    return None



import logging
import json

from deltacat.storage import PartitionLocator
from deltacat.compute.compactor import RoundCompletionInfo
from deltacat import logs

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def get_round_completion_file_s3_url(
        bucket: str,
        source_partition_locator: PartitionLocator,
        pki_root_path: str) -> str:

    base_url = source_partition_locator.path(f"s3://{bucket}")
    return f"{base_url}/{pki_root_path}.json"


def read_round_completion_file(
        bucket: str,
        source_partition_locator: PartitionLocator,
        primary_key_index_root_path: str) -> RoundCompletionInfo:

    from deltacat.aws import s3u as s3_utils
    round_completion_file_url = get_round_completion_file_s3_url(
        bucket,
        source_partition_locator,
        primary_key_index_root_path,
    )
    logger.info(
        f"reading round completion file from: {round_completion_file_url}")
    round_completion_info = None
    result = s3_utils.download(round_completion_file_url, False)
    if result:
        json_str = result["Body"].read().decode("utf-8")
        round_completion_info = RoundCompletionInfo(json.loads(json_str))
        logger.info(f"read round completion info: {round_completion_info}")
    return round_completion_info


def write_round_completion_file(
        bucket: str,
        source_partition_locator: PartitionLocator,
        primary_key_index_root_path: str,
        round_completion_info: RoundCompletionInfo):

    from deltacat.aws import s3u as s3_utils
    logger.info(
        f"writing round completion file contents: {round_completion_info}")
    round_completion_file_s3_url = get_round_completion_file_s3_url(
        bucket,
        source_partition_locator,
        primary_key_index_root_path,
    )
    logger.info(
        f"writing round completion file to: {round_completion_file_s3_url}")
    s3_utils.upload(
        round_completion_file_s3_url,
        str(json.dumps(round_completion_info))
    )
    logger.info(
        f"round completion file written to: {round_completion_file_s3_url}")



import os
import logging
import pathlib
from logging import handlers, Logger, Handler, FileHandler
from deltacat.constants import DELTACAT_LOG_LEVEL, APPLICATION_LOG_LEVEL

DEFAULT_LOG_LEVEL = "INFO"
DEFAULT_LOG_FORMAT = \
    "%(asctime)s\t%(levelname)s %(filename)s:%(lineno)s -- %(message)s"
DEFAULT_MAX_BYTES_PER_LOG = 2 ^ 20 * 256  # 256 MiB
DEFAULT_BACKUP_COUNT = 0

DEFAULT_APPLICATION_LOG_DIR = "/tmp/deltacat/var/output/logs/"
DEFAULT_APPLICATION_LOG_BASE_FILE_NAME = "application.info.log"
DEFAULT_DEBUG_APPLICATION_LOG_BASE_FILE_NAME = "application.debug.log"

DEFAULT_DELTACAT_LOG_DIR = "/tmp/deltacat/var/output/logs/"
DEFAULT_DELTACAT_LOG_BASE_FILE_NAME = "deltacat-python.info.log"
DEFAULT_DEBUG_DELTACAT_LOG_BASE_FILE_NAME = "deltacat-python.debug.log"


def _add_logger_handler(
        logger: Logger,
        handler: Handler) -> Logger:

    logger.setLevel(logging.getLevelName("DEBUG"))
    logger.addHandler(handler)
    return logger


def _create_rotating_file_handler(
        log_directory: str,
        log_base_file_name: str,
        logging_level: str = DEFAULT_LOG_LEVEL,
        max_bytes_per_log_file: int = DEFAULT_MAX_BYTES_PER_LOG,
        backup_count: int = DEFAULT_BACKUP_COUNT,
        logging_format: str = DEFAULT_LOG_FORMAT) -> FileHandler:

    if type(logging_level) is str:
        logging_level = logging.getLevelName(logging_level.upper())
    assert log_base_file_name, "log file name is required"
    assert log_directory, "log directory is required"

    log_dir_path = pathlib.Path(log_directory)
    log_dir_path.mkdir(parents=True, exist_ok=True)
    handler = logging.handlers.RotatingFileHandler(
        os.path.join(log_directory, log_base_file_name),
        maxBytes=max_bytes_per_log_file,
        backupCount=backup_count)
    handler.setFormatter(logging.Formatter(logging_format))
    handler.setLevel(logging_level)
    return handler


def _file_handler_exists(
        logger: Logger,
        log_dir: str,
        log_base_file_name: str) -> bool:

    handler_exists = False
    base_file_path = os.path.join(log_dir, log_base_file_name)
    if len(logger.handlers) > 0:
        norm_base_file_path = os.path.normpath(base_file_path)
        handler_exists = any([
            isinstance(handler, logging.FileHandler)
            and os.path.normpath(handler.baseFilename) == norm_base_file_path
            for handler in logger.handlers
        ])
    return handler_exists


def _configure_logger(
        logger: Logger,
        log_level: str,
        log_dir: str,
        log_base_file_name: str,
        debug_log_base_file_name: str) -> Logger:

    primary_log_level = log_level
    logger.propagate = False
    if log_level.upper() == "DEBUG":
        if not _file_handler_exists(logger, log_dir, debug_log_base_file_name):
            handler = _create_rotating_file_handler(
                log_dir,
                debug_log_base_file_name,
                "DEBUG",
            )
            _add_logger_handler(logger, handler)
            primary_log_level = "INFO"
    if not _file_handler_exists(logger, log_dir, log_base_file_name):
        handler = _create_rotating_file_handler(
            log_dir,
            log_base_file_name,
            primary_log_level,
        )
        _add_logger_handler(logger, handler)
    return logger


def configure_deltacat_logger(logger: Logger) -> Logger:
    return _configure_logger(
        logger,
        DELTACAT_LOG_LEVEL,
        DEFAULT_DELTACAT_LOG_DIR,
        DEFAULT_DELTACAT_LOG_BASE_FILE_NAME,
        DEFAULT_DEBUG_DELTACAT_LOG_BASE_FILE_NAME
    )


def configure_application_logger(logger: Logger) -> Logger:
    return _configure_logger(
        logger,
        APPLICATION_LOG_LEVEL,
        DEFAULT_APPLICATION_LOG_DIR,
        DEFAULT_APPLICATION_LOG_BASE_FILE_NAME,
        DEFAULT_DEBUG_APPLICATION_LOG_BASE_FILE_NAME
    )



import ray
import re
import time
import yaml
import logging
from typing import Optional, Union, List, Dict, Any, Callable, Tuple
from ray.util.placement_group import (
	placement_group,
	placement_group_table,
	get_current_placement_group
)

from ray.util.scheduling_strategies import PlacementGroupSchedulingStrategy
from ray.experimental.state.api import get_node, get_placement_group


from deltacat import logs
logger = logs.configure_deltacat_logger(logging.getLogger(__name__))

#Limitation of current node group or placement group manager
#Must run on driver or head node bc state.api needs to query dashboard api server at 127.0.0.1.
#Issue: https://github.com/ray-project/ray/issues/29959

class NodeGroupManager():

	def __init__(self,path: str, gname: str):
		"""Node Group Manager
		Args:
			path: cluster yaml file
			gname: node group prefix, e.g., 'partition'
		"""
		#cluster init status:
		self.NODE_GROUP_PREFIX=gname
		self.cluster_config=self._read_yaml(path)
		self.init_groups = self._cluster_node_groups(self.cluster_config)
		self.init_group_res = self._parse_node_resources()
		
	def _cluster_node_groups(self, config: Dict[str, Any]) -> Dict[str, Any]:
		"""Get Worker Groups
		Args:
			config: cluster yaml data
		Returns: 
			worker groups: a dict of worker node group

		"""
		avail_node_types =  list(config['available_node_types'].items())
		#exclude head node type
		head_node_types = [nt for nt in avail_node_types if 'resources' in nt[1] and 'CPU' in nt[1]['resources'] and nt[1]['resources']['CPU']==0][0]
		worker_node_types = [x for x in avail_node_types if x !=head_node_types]
		#assuming homogenous cluster
		#in future, update with fleet resource
		if len(worker_node_types)>0:
			self.INSTANCE_TYPE = worker_node_types[0][1]['node_config']['InstanceType']
		return worker_node_types


	def _read_yaml(self, path: str) -> Dict[str, Any]:
		with open(path, "rt") as f:
			return yaml.safe_load(f)

	def _update_groups(self) -> List[Tuple[str, float]]:
		"""
			Node groups can come and go during runtime, whenever a node group is needed, we need to check the current available groups
		Returns:
			current_groups: dict of custom resource groups
		"""
		#Add 1.1 second latency to avoid inconsistency issue between raylet and head
		time.sleep(1.1)
		all_available_res = ray.available_resources()
		current_groups =[(k,all_available_res[k]) for k in all_available_res.keys() if self.NODE_GROUP_PREFIX in k]
		return current_groups

	def _parse_node_resources(self) -> Dict[str, Dict[str, float]]:
		"""
			Parse resources per node to get detailed resource tighted to each node group
			Returns:
				group_res: a dict of resources, e.g., {'CPU':0,'memory':0,'object_store_memory':0}
		"""
		all_available_resources= ray._private.state.state._available_resources_per_node()
		group_keys = [x[0] for x in self.init_groups]
		group_res={}
		for k in group_keys:
			group_res[k]={'CPU':0,'memory':0,'object_store_memory':0,'node_id':[]}
		for v in all_available_resources.values():
			keys =v.keys()
			r = re.compile(self.NODE_GROUP_PREFIX)
			partition=list(filter(r.match, list(keys)))
			r = re.compile("node:")
			node_id = list(filter(r.match, list(keys)))
			if len(partition)>0:
				partition = partition[0]
			if len(node_id)>0:
				node_id = node_id[0]
			if self.NODE_GROUP_PREFIX in partition:
				group_res[partition]['CPU']+=v['CPU']
				group_res[partition]['memory']+=v['memory']
				group_res[partition]['object_store_memory']+=v['object_store_memory']
				group_res[partition]['node_id'].append(node_id)
		return group_res

	def _update_group_res(self, gname: str) -> Dict[str, Union[str, float]]:
		"""
			Get the realtime resource of a node group
			Args:
				gname: name of node group
			Returns:
				group_res: dict of updated resource(cpu, memory, object store memory) for a given group
		"""
		all_available_resources= ray._private.state.state._available_resources_per_node()
		group_res={'CPU':0,'memory':0,'object_store_memory':0,'node_id':[]}
		for v in all_available_resources.values():
			keys =v.keys()
			r = re.compile("node:")
			node_id = list(filter(r.match, list(keys)))
			if len(node_id)>0:
				node_id = node_id[0]
			if gname in v.keys():
				group_res['CPU']+=v['CPU']
				group_res['memory']+=v['memory']
				group_res['object_store_memory']+=v['object_store_memory']
				group_res['node_id'].append(node_id)
		return group_res

	def get_one_group(self) -> Optional[Dict[str, Union[str, float]]]:
		"""
			Pop up one node group 
			Returns:
				group_res: dict of node group resource, {"group":"partition_1","CPU":2,...}
		"""
		current_groups = self._update_groups()
		if len(current_groups)>0:
			gname = current_groups[-1][0]
			group_res=self._update_group_res(gname)
			group_res['group']=gname
			try:
				group_res['group_res']=ray.available_resources()[gname]
			except Exception as e:
				logger.info(f"There is no available resources for {gname}")
				return None
			return group_res
		else:
			return None

	def get_group_by_name(self, gname: str) -> Optional[Dict[str, Union[str, float]]]:
		"""
			Get the specific node group given its pre-filled name
			Args: 
				gname: name of the node group
			Returns:
				group_res: dict of node group resource

		"""
		group_res=self._update_group_res(gname)
		group_res['group']=gname		
		try:
			group_res['group_res']=ray.available_resources()[gname]
		except Exception as e:
			logger.info(f"There is no available resources for {gname}")
			return None
		return group_res


class PlacementGroupManager():
	"""Placement Group Manager
	Create a list of placement group with the desired number of cpus
	e.g., create a pg with 32 cpus, then this class will look for a node that has 32 cpus, and collect all 
	resources, including cpu, memory, and object store;
	How to use:
		```
			from deltacat.utils.placement import PlacementGroupManager as pgm
			pgm = pgm(10, 32)
			pg_configs = pgm.pgs
			opts = pg_configs[0][0]
			fun.options(**opts).remote()
		```
	Args:
		num_pgs: number of placement groups to be created
		instance_cpus: number of cpus per instance
	"""
	def __init__(self, num_pgs: int, instance_cpus: int, instance_type: int = 8, time_out: Optional[float] = None):
		head_res_key = self.get_current_node_resource_key()
		self._pg_configs = ray.get([_config.options(resources={head_res_key:0.01}).remote(instance_cpus, instance_type) for _ in range(num_pgs)])
	@property
	def pgs(self):
		return self._pg_configs

	def get_current_node_resource_key(self) -> str: 
	    current_node_id = ray.get_runtime_context().node_id.hex() 
	    for node in ray.nodes(): 
	        if node["NodeID"] == current_node_id: 
	            # Found the node. 
	            for key in node["Resources"].keys(): 
	                if key.startswith("node:"): 
	                    return key
@ray.remote(num_cpus=0.01)
def _config(instance_cpus: int, instance_type: int, time_out: Optional[float] = None) -> Tuple[Dict[str,Any], Dict[str,Any]]:
	pg_config = None
	try:
		opts ={}
		cluster_resources={}
		num_bundles = (int)(instance_cpus/instance_type)
		bundles = [{'CPU':instance_type} for _ in range(num_bundles)]
		pg = placement_group(bundles, strategy="SPREAD")
		ray.get(pg.ready(), timeout=time_out)
		if not pg:
			return None
		opts = {"scheduling_strategy":PlacementGroupSchedulingStrategy(
			placement_group=pg, placement_group_capture_child_tasks=True)
		}
		pg_id = placement_group_table(pg)['placement_group_id']
		pg_details = get_placement_group(pg_id)
		bundles = pg_details['bundles']
		node_ids =[]
		for bd in bundles:
			node_ids.append(bd['node_id'])
		#query available resources given list of node id
		all_nodes_available_res = ray._private.state.state._available_resources_per_node()
		pg_res = {'CPU':0,'memory':0,'object_store_memory':0,'node_id':[]}
		for node_id in node_ids:
			if node_id in all_nodes_available_res:
				v = all_nodes_available_res[node_id]
				node_detail = get_node(node_id)
				pg_res['CPU']+=node_detail['resources_total']['CPU']
				pg_res['memory']+=v['memory']
				pg_res['object_store_memory']+=v['object_store_memory']
				pg_res['node_id'].append(node_id)
		cluster_resources['CPU'] = int(pg_res['CPU'])
		cluster_resources['memory'] = float(pg_res['memory'])
		cluster_resources['object_store_memory'] = float(pg_res['object_store_memory'])
		cluster_resources['node_id'] = pg_res['node_id']
		pg_config=[opts,cluster_resources]
		logger.info(f"pg has resources:{cluster_resources}")

	except Exception as e:
		logger.error(f"placement group error:{e}")
		pass
	return pg_config




import pandas as pd
import csv
import math
import io
import logging
import pyarrow as pa

from fsspec import AbstractFileSystem

from ray.data.datasource import BlockWritePathProvider

from deltacat.types.media import ContentType, ContentEncoding, \
    EXPLICIT_COMPRESSION_CONTENT_TYPES
from deltacat.types.media import DELIMITED_TEXT_CONTENT_TYPES, \
    TABULAR_CONTENT_TYPES
from deltacat import logs
from deltacat.utils.common import ReadKwargsProvider, ContentTypeKwargsProvider
from deltacat.utils.performance import timed_invocation

from typing import Any, Callable, Dict, Iterable, List, Optional

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


CONTENT_TYPE_TO_PD_READ_FUNC: Dict[str, Callable] = {
    ContentType.UNESCAPED_TSV.value: pd.read_csv,
    ContentType.TSV.value: pd.read_csv,
    ContentType.CSV.value: pd.read_csv,
    ContentType.PSV.value: pd.read_csv,
    ContentType.PARQUET.value: pd.read_parquet,
    ContentType.FEATHER.value: pd.read_feather,
    ContentType.ORC.value: pd.read_orc,
    ContentType.JSON.value: pd.read_json
}


class ReadKwargsProviderPandasCsvPureUtf8(ContentTypeKwargsProvider):
    """ReadKwargsProvider impl that reads columns of delimited text files
    as UTF-8 strings (i.e. disables type inference). Useful for ensuring
    lossless reads of UTF-8 delimited text datasets and improving read
    performance in cases where type casting is not required."""
    def __init__(self, include_columns: Optional[Iterable[str]] = None):
        self.include_columns = include_columns

    def _get_kwargs(
            self,
            content_type: str,
            kwargs: Dict[str, Any]) -> Dict[str, Any]:
        if content_type in DELIMITED_TEXT_CONTENT_TYPES:
            include_columns = self.include_columns \
                if self.include_columns else kwargs.get("usecols")
            if not include_columns:
                # read all columns as strings
                kwargs["dtype"] = str
            else:
                # read only the included columns as strings
                kwargs["dtype"] = {
                    column_name: str for column_name in include_columns
                }
            # use the fastest available engine for pure utf-8 reads
            kwargs["engine"] = "pyarrow"
        return kwargs


def content_type_to_reader_kwargs(content_type: str) -> Dict[str, Any]:
    if content_type == ContentType.UNESCAPED_TSV.value:
        return {
            "sep": "\t",
            "header": None,
            "na_values": [""],
            "keep_default_na": False,
        }
    if content_type == ContentType.TSV.value:
        return {
            "sep": "\t",
            "header": None
        }
    if content_type == ContentType.CSV.value:
        return {
            "sep": ",",
            "header": None
        }
    if content_type == ContentType.PSV.value:
        return {
            "sep": "|",
            "header": None
        }
    if content_type in {ContentType.PARQUET.value,
                        ContentType.FEATHER.value,
                        ContentType.ORC.value,
                        ContentType.JSON.value}:
        return {}
    raise ValueError(f"Unsupported content type: {content_type}")


ENCODING_TO_PD_COMPRESSION: Dict[str, str] = {
    ContentEncoding.GZIP.value: "gzip",
    ContentEncoding.BZIP2.value: "bz2",
    ContentEncoding.IDENTITY.value: "none"
}


def slice_dataframe(
        dataframe: pd.DataFrame,
        max_len: Optional[int]) -> List[pd.DataFrame]:
    """
    Iteratively create dataframe slices.
    """
    if max_len is None:
        return [dataframe]
    dataframes = []
    num_slices = math.ceil(len(dataframe) / max_len)
    for i in range(num_slices):
        dataframes.append(dataframe[i * max_len: (i + 1) * max_len])
    return dataframes


def concat_dataframes(dataframes: List[pd.DataFrame]) \
        -> Optional[pd.DataFrame]:
    if dataframes is None or not len(dataframes):
        return None
    if len(dataframes) == 1:
        return next(iter(dataframes))
    return pd.concat(dataframes, axis=0, copy=False)


def _add_column_kwargs(
        content_type: str,
        column_names: Optional[List[str]],
        include_columns: Optional[List[str]],
        kwargs: Dict[str, Any]):

    if content_type in DELIMITED_TEXT_CONTENT_TYPES:
        kwargs["names"] = column_names
        kwargs["usecols"] = include_columns
    else:
        if content_type in TABULAR_CONTENT_TYPES:
            kwargs["columns"]: include_columns
        else:
            if include_columns:
                logger.warning(
                    f"Ignoring request to include columns {include_columns} "
                    f"for non-tabular content type {content_type}")


def s3_file_to_dataframe(
        s3_url: str,
        content_type: str,
        content_encoding: str,
        column_names: Optional[List[str]] = None,
        include_columns: Optional[List[str]] = None,
        pd_read_func_kwargs_provider: Optional[ReadKwargsProvider] = None,
        **s3_client_kwargs) -> pd.DataFrame:

    from deltacat.aws import s3u as s3_utils
    logger.debug(f"Reading {s3_url} to Pandas. Content type: {content_type}. "
                 f"Encoding: {content_encoding}")
    s3_obj = s3_utils.get_object_at_url(
        s3_url,
        **s3_client_kwargs
    )
    logger.debug(f"Read S3 object from {s3_url}: {s3_obj}")
    pd_read_func = CONTENT_TYPE_TO_PD_READ_FUNC[content_type]
    args = [io.BytesIO(s3_obj["Body"].read())]
    kwargs = content_type_to_reader_kwargs(content_type)
    _add_column_kwargs(content_type, column_names, include_columns, kwargs)

    if content_type in EXPLICIT_COMPRESSION_CONTENT_TYPES:
        kwargs["compression"] = ENCODING_TO_PD_COMPRESSION.get(
            content_encoding,
            "infer"
        )
    if pd_read_func_kwargs_provider:
        kwargs = pd_read_func_kwargs_provider(content_type, kwargs)
    logger.debug(f"Reading {s3_url} via {pd_read_func} with kwargs: {kwargs}")
    dataframe, latency = timed_invocation(
        pd_read_func,
        *args,
        **kwargs
    )
    logger.debug(f"Time to read {s3_url} into Pandas Dataframe: {latency}s")
    return dataframe


def dataframe_size(dataframe: pd.DataFrame) -> int:
    # TODO (pdames): inspect latency vs. deep memory usage inspection
    return int(dataframe.memory_usage().sum())


def write_csv(
        dataframe: pd.DataFrame,
        path: str,
        *,
        filesystem: AbstractFileSystem,
        **kwargs) -> None:
    with filesystem.open(path, "wb") as f:
        # TODO (pdames): Add support for client-specified compression types.
        with pa.CompressedOutputStream(f, ContentEncoding.GZIP.value) as out:
            dataframe.to_csv(out, **kwargs)


def write_parquet(
        dataframe: pd.DataFrame,
        path: str,
        *,
        filesystem: AbstractFileSystem,
        **kwargs) -> None:
    with filesystem.open(path, "wb") as f:
        dataframe.to_parquet(f, **kwargs)


def write_feather(
        dataframe: pd.DataFrame,
        path: str,
        *,
        filesystem: AbstractFileSystem,
        **kwargs) -> None:
    with filesystem.open(path, "wb") as f:
        dataframe.to_feather(f, **kwargs)


def write_json(
        dataframe: pd.DataFrame,
        path: str,
        *,
        filesystem: AbstractFileSystem,
        **kwargs) -> None:
    with filesystem.open(path, "wb") as f:
        # TODO (pdames): Add support for client-specified compression types.
        with pa.CompressedOutputStream(f, ContentEncoding.GZIP.value) as out:
            dataframe.to_json(out, **kwargs)


CONTENT_TYPE_TO_PD_WRITE_FUNC: Dict[str, Callable] = {
    ContentType.UNESCAPED_TSV: write_csv,
    ContentType.TSV.value: write_csv,
    ContentType.CSV.value: write_csv,
    ContentType.PSV.value: write_csv,
    ContentType.PARQUET.value: write_parquet,
    ContentType.FEATHER.value: write_feather,
    ContentType.JSON.value: write_json,
}


def content_type_to_writer_kwargs(content_type: str) -> Dict[str, Any]:
    if content_type == ContentType.UNESCAPED_TSV.value:
        return {
            "sep": "\t",
            "header": False,
            "na_rep": [""],
            "line_terminator": "\n",
            "quoting": csv.QUOTE_NONE,
            "index": False,
        }
    if content_type == ContentType.TSV.value:
        return {
            "sep": "\t",
            "header": False,
            "line_terminator": "\n",
            "index": False,
        }
    if content_type == ContentType.CSV.value:
        return {
            "sep": ",",
            "header": False,
            "line_terminator": "\n",
            "index": False,
        }
    if content_type == ContentType.PSV.value:
        return {
            "sep": "|",
            "header": False,
            "line_terminator": "\n",
            "index": False,
        }
    if content_type == ContentType.PARQUET.value:
        return {"index": False}
    if content_type == ContentType.FEATHER.value:
        return {}
    if content_type == ContentType.JSON.value:
        return {"index": False}
    raise ValueError(f"Unsupported content type: {content_type}")


def dataframe_to_file(
        dataframe: pd.DataFrame,
        base_path: str,
        file_system: AbstractFileSystem,
        block_path_provider: BlockWritePathProvider,
        content_type: str = ContentType.PARQUET.value,
        **kwargs) -> None:
    """
    Writes the given Pandas Dataframe to a file.
    """
    writer = CONTENT_TYPE_TO_PD_WRITE_FUNC.get(content_type)
    writer_kwargs = content_type_to_writer_kwargs(content_type)
    writer_kwargs.update(kwargs)
    if not writer:
        raise NotImplementedError(
            f"Pandas writer for content type '{content_type}' not "
            f"implemented. Known content types: "
            f"{CONTENT_TYPE_TO_PD_WRITE_FUNC.keys}")
    path = block_path_provider(base_path)
    writer(
        dataframe,
        path,
        filesystem=file_system,
        **writer_kwargs
    )



import logging
import ray
import pyarrow as pa
import numpy as np

from ray.types import ObjectRef

from deltacat import logs
from deltacat.compute.compactor import PrimaryKeyIndexVersionLocator
from deltacat.compute.compactor.utils import primary_key_index as pki

from typing import List, Tuple

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def group_file_records_by_pk_hash_bucket(
        pki_table: pa.Table,
        num_buckets: int) -> np.ndarray:
    # generate the new table for each new hash bucket
    hash_bucket_to_indices = pki.group_record_indices_by_hash_bucket(
        pki_table,
        num_buckets,
    )
    hash_bucket_to_table = np.empty([num_buckets], dtype="object")
    for hash_bucket, indices in enumerate(hash_bucket_to_indices):
        if indices:
            hash_bucket_to_table[hash_bucket] = pki_table.take(indices)
    return hash_bucket_to_table


@ray.remote(num_cpus=1,num_returns=2)
def rehash_bucket(
        hash_bucket_index: int,
        s3_bucket: str,
        old_pki_version_locator: PrimaryKeyIndexVersionLocator,
        num_buckets: int,
        num_groups: int) -> Tuple[np.ndarray, List[ObjectRef]]:

    logger.info(f"Starting rehash bucket task...")
    tables = pki.download_hash_bucket_entries(
        s3_bucket,
        hash_bucket_index,
        old_pki_version_locator,
    )
    prior_pk_index_table = pa.concat_tables(tables)
    hash_bucket_to_table = group_file_records_by_pk_hash_bucket(
        prior_pk_index_table,
        num_buckets,
    )
    hash_bucket_group_to_obj_id, object_refs = pki.group_hash_bucket_indices(
        hash_bucket_to_table,
        num_buckets,
        num_groups,
    )
    logger.info(f"Finished rehash bucket task...")
    return hash_bucket_group_to_obj_id, object_refs



import time
from deltacat.utils.ray_utils.collections import DistributedCounter
from typing import Any, Callable, Tuple


def invoke_with_perf_counter(
        counter: DistributedCounter,
        counter_key: Any,
        func: Callable,
        *args,
        **kwargs) -> Tuple[Any, float]:

    start = time.perf_counter()
    result = func(*args, **kwargs)
    stop = time.perf_counter()
    latency = stop - start
    counter.increment.remote(
        counter_key,
        latency)
    return result, latency
